"""Exceptions for tmuxp."""

from __future__ import annotations

import logging

from libtmux._internal.query_list import ObjectDoesNotExist

from ._compat import implements_to_string

logger = logging.getLogger(__name__)


class TmuxpException(Exception):
    """Base Exception for Tmuxp Errors."""


class WorkspaceError(TmuxpException):
    """Error parsing tmuxp workspace data."""


class SessionNotFound(TmuxpException):
    """tmux session not found."""

    def __init__(
        self,
        session_target: str | None = None,
        *args: object,
        **kwargs: object,
    ) -> None:
        msg = "Session not found"
        if session_target is not None:
            msg += f": {session_target}"
        return super().__init__(msg, *args, **kwargs)


class WindowNotFound(TmuxpException):
    """tmux window not found."""

    def __init__(
        self,
        window_target: str | None = None,
        *args: object,
        **kwargs: object,
    ) -> None:
        msg = "Window not found"
        if window_target is not None:
            msg += f": {window_target}"
        return super().__init__(msg, *args, **kwargs)


class PaneNotFound(TmuxpException):
    """tmux pane not found."""

    def __init__(
        self,
        pane_target: str | None = None,
        *args: object,
        **kwargs: object,
    ) -> None:
        msg = "Pane not found"
        if pane_target is not None:
            msg += f": {pane_target}"
        return super().__init__(msg, *args, **kwargs)


class EmptyWorkspaceException(WorkspaceError):
    """Workspace file is empty."""

    def __init__(
        self,
        *args: object,
        **kwargs: object,
    ) -> None:
        return super().__init__("Session configuration is empty.", *args, **kwargs)


class SessionMissingWorkspaceException(WorkspaceError, ObjectDoesNotExist):
    """Session missing while loading tmuxp workspace."""

    def __init__(self, *args: object, **kwargs: object) -> None:
        return super().__init__(
            "No session object exists for WorkspaceBuilder. "
            "Tip: Add session_name in constructor or run WorkspaceBuilder.build()",
            *args,
            **kwargs,
        )


class ActiveSessionMissingWorkspaceException(WorkspaceError):
    """Active session cannot be found while loading tmuxp workspace."""

    def __init__(self, *args: object, **kwargs: object) -> None:
        return super().__init__("No session active.", *args, **kwargs)


class WorkspaceBuilderError(WorkspaceError):
    """Base error for resolving and validating a workspace builder."""


class WorkspaceBuilderNotFound(WorkspaceBuilderError):
    """Configured ``workspace_builder`` could not be resolved.

    >>> print(WorkspaceBuilderNotFound("nope"))
    Workspace builder 'nope' could not be found....
    """

    def __init__(
        self,
        target: str,
        available: list[str] | None = None,
        *args: object,
        **kwargs: object,
    ) -> None:
        msg = f"Workspace builder {target!r} could not be found."
        if available:
            names = ", ".join(sorted(available))
            msg += f" Available entry point builders: {names}."
        else:
            msg += (
                " Provide a Python dotted path (e.g. 'package.module:Builder') "
                "or register an entry point in the 'tmuxp.workspace_builders' group."
            )
        return super().__init__(msg, *args, **kwargs)


class WorkspaceBuilderImportError(WorkspaceBuilderError):
    """Configured ``workspace_builder`` failed to import.

    >>> print(WorkspaceBuilderImportError("pkg:B"))
    Could not import workspace builder 'pkg:B'....
    """

    def __init__(
        self,
        target: str,
        reason: str | None = None,
        paths: list[str] | None = None,
        *args: object,
        **kwargs: object,
    ) -> None:
        msg = f"Could not import workspace builder {target!r}."
        if reason:
            msg += f" {reason}"
        if paths:
            msg += f" Searched trusted paths: {', '.join(paths)}."
        msg += (
            " Confirm the builder is importable, or add its directory to "
            "'workspace_builder_paths'."
        )
        return super().__init__(msg, *args, **kwargs)


class InvalidWorkspaceBuilder(WorkspaceBuilderError):
    """Resolved ``workspace_builder`` object is not a usable builder.

    >>> print(InvalidWorkspaceBuilder("pkg:B"))
    'pkg:B' is not a valid workspace builder.
    """

    def __init__(
        self,
        target: str,
        reason: str | None = None,
        *args: object,
        **kwargs: object,
    ) -> None:
        msg = f"{target!r} is not a valid workspace builder"
        msg += f": {reason}" if reason else "."
        return super().__init__(msg, *args, **kwargs)


class WorkspaceBuilderPathError(WorkspaceBuilderError):
    """A ``workspace_builder_paths`` entry is not a usable directory.

    >>> print(WorkspaceBuilderPathError("/x"))
    workspace_builder_paths entry is invalid: /x....
    """

    def __init__(
        self,
        path: str,
        reason: str | None = None,
        *args: object,
        **kwargs: object,
    ) -> None:
        msg = f"workspace_builder_paths entry is invalid: {path}."
        msg += f" {reason}" if reason else " Each entry must be an existing directory."
        return super().__init__(msg, *args, **kwargs)


class InvalidWorkspaceBuilderOption(WorkspaceBuilderError):
    """A ``workspace_builder_options`` value is invalid.

    >>> print(InvalidWorkspaceBuilderOption("bad"))
    Invalid workspace_builder_options: bad
    """

    def __init__(self, reason: str, *args: object, **kwargs: object) -> None:
        return super().__init__(
            f"Invalid workspace_builder_options: {reason}",
            *args,
            **kwargs,
        )


class TmuxpPluginException(TmuxpException):
    """Base Exception for Tmuxp Errors."""


class BeforeLoadScriptNotExists(OSError):
    """Raises if shell script could not be found."""

    def __init__(self, *args: object, **kwargs: object) -> None:
        super().__init__(*args, **kwargs)

        self.strerror = f"before_script file '{self.strerror}' doesn't exist."


@implements_to_string
class BeforeLoadScriptError(Exception):
    """Shell script execution error.

    Replaces :class:`subprocess.CalledProcessError` for
    :func:`tmuxp.util.run_before_script`.
    """

    def __init__(
        self,
        returncode: int,
        cmd: str,
        output: str | None = None,
    ) -> None:
        self.returncode = returncode
        self.cmd = cmd
        self.output = output
        self.message = (
            f"before_script failed with returncode {self.returncode}.\n"
            f"command: {self.cmd}\n"
            "Error output:\n"
            f"{self.output}"
        )

    def __str__(self) -> str:
        """Return shell error message."""
        return self.message



"""Fixtures for tmuxp plugins for tmux version exceptions."""

from __future__ import annotations

import typing as t

from .test_plugin_helpers import MyTestTmuxpPlugin

if t.TYPE_CHECKING:
    from ._types import PluginTestConfigSchema


class TmuxVersionFailMinPlugin(MyTestTmuxpPlugin):
    """Tmuxp plugin that fails when tmux below minimum version constraint."""

    def __init__(self) -> None:
        config: PluginTestConfigSchema = {
            "plugin_name": "tmux-min-version-fail",
            "tmux_min_version": "1.8",
            "tmux_version": "1.7",
        }
        MyTestTmuxpPlugin.__init__(self, **config)


class TmuxVersionFailMaxPlugin(MyTestTmuxpPlugin):
    """Tmuxp plugin that fails when tmux above maximum version constraint."""

    def __init__(self) -> None:
        config: PluginTestConfigSchema = {
            "plugin_name": "tmux-max-version-fail",
            "tmux_max_version": "3.0",
            "tmux_version": "3.5",
        }
        MyTestTmuxpPlugin.__init__(self, **config)


class TmuxVersionFailIncompatiblePlugin(MyTestTmuxpPlugin):
    """Tmuxp plugin that fails when tmux version constraint is invalid."""

    def __init__(self) -> None:
        config: PluginTestConfigSchema = {
            "plugin_name": "tmux-incompatible-version-fail",
            "tmux_version_incompatible": ["2.3"],
            "tmux_version": "2.3",
        }

        MyTestTmuxpPlugin.__init__(self, **config)



"""Utility functions for tmuxp fixtures."""

from __future__ import annotations

import pathlib

from tests.constants import FIXTURE_PATH


def get_workspace_file(
    file: str | pathlib.Path,
) -> pathlib.Path:
    """Return fixture data, relative to __file__."""
    if isinstance(file, str):
        file = pathlib.Path(file)

    return FIXTURE_PATH / file


def read_workspace_file(
    file: pathlib.Path | str,
) -> str:
    """Return fixture data, relative to __file__."""
    if isinstance(file, str):
        file = pathlib.Path(file)

    return get_workspace_file(file).open().read()


def write_config(
    config_path: pathlib.Path,
    filename: str,
    content: str,
) -> pathlib.Path:
    """Write configuration content to file."""
    config = config_path / filename
    config.write_text(content, encoding="utf-8")
    return config



"""Conftest.py (root-level).

We keep this in root pytest fixtures in pytest's doctest plugin to be available, as well
as avoiding conftest.py from being included in the wheel, in addition to pytest_plugin
for pytester only being available via the root directory.

See "pytest_plugins in non-top-level conftest files" in
https://docs.pytest.org/en/stable/deprecations.html
"""

from __future__ import annotations

import logging
import os
import pathlib
import shutil
import typing as t

import pytest
from _pytest.doctest import DoctestItem
from libtmux.test.random import namer

from tests.fixtures import utils as test_utils
from tmuxp.workspace.finders import get_workspace_dir

if t.TYPE_CHECKING:
    from libtmux.session import Session

logger = logging.getLogger(__name__)
USING_ZSH = "zsh" in os.getenv("SHELL", "")


@pytest.fixture(autouse=True)
def _pin_test_shell_env(monkeypatch: pytest.MonkeyPatch) -> None:
    """Pin ``$SHELL`` to ``/bin/sh`` for every test.

    tmux falls back to ``$SHELL`` for new panes when ``default-shell`` is
    unset, so pinning the env var here forces every test pane to spawn
    ``/bin/sh`` instead of the contributor's interactive shell. Eliminates
    rcfile init cost (zsh / oh-my-zsh / bash profile chains) and yields
    deterministic prompt output that doesn't flake capture-pane assertions.
    ``monkeypatch`` restores the prior value at teardown.
    """
    monkeypatch.setenv("SHELL", "/bin/sh")


@pytest.fixture(autouse=USING_ZSH, scope="session")
def zshrc(user_path: pathlib.Path) -> pathlib.Path | None:
    """Quiets ZSH default message.

    Needs a startup file .zshenv, .zprofile, .zshrc, .zlogin.
    """
    if not USING_ZSH:
        return None
    p = user_path / ".zshrc"
    p.touch()
    return p


@pytest.fixture(autouse=True)
def home_path_default(monkeypatch: pytest.MonkeyPatch, user_path: pathlib.Path) -> None:
    """Set HOME to user_path (random, temporary directory)."""
    monkeypatch.setenv("HOME", str(user_path))


@pytest.fixture
def tmuxp_configdir(user_path: pathlib.Path) -> pathlib.Path:
    """Ensure and return tmuxp config directory."""
    xdg_config_dir = user_path / ".config"
    xdg_config_dir.mkdir(exist_ok=True)

    tmuxp_configdir = xdg_config_dir / "tmuxp"
    tmuxp_configdir.mkdir(exist_ok=True)
    return tmuxp_configdir


@pytest.fixture
def tmuxp_configdir_default(
    monkeypatch: pytest.MonkeyPatch,
    tmuxp_configdir: pathlib.Path,
) -> None:
    """Set tmuxp configuration directory for ``TMUXP_CONFIGDIR``."""
    monkeypatch.setenv("TMUXP_CONFIGDIR", str(tmuxp_configdir))
    assert get_workspace_dir() == str(tmuxp_configdir)


@pytest.fixture
def monkeypatch_plugin_test_packages(monkeypatch: pytest.MonkeyPatch) -> None:
    """Monkeypatch tmuxp plugin fixtures to python path."""
    paths = [
        "tests/fixtures/pluginsystem/plugins/tmuxp_test_plugin_bwb/",
        "tests/fixtures/pluginsystem/plugins/tmuxp_test_plugin_bs/",
        "tests/fixtures/pluginsystem/plugins/tmuxp_test_plugin_r/",
        "tests/fixtures/pluginsystem/plugins/tmuxp_test_plugin_owc/",
        "tests/fixtures/pluginsystem/plugins/tmuxp_test_plugin_awf/",
        "tests/fixtures/pluginsystem/plugins/tmuxp_test_plugin_fail/",
    ]
    for path in paths:
        monkeypatch.syspath_prepend(str(pathlib.Path(path).resolve()))


@pytest.fixture
def session_params(session_params: dict[str, t.Any]) -> dict[str, t.Any]:
    """Terminal-friendly tmuxp session_params for dimensions."""
    session_params.update({"x": 800, "y": 600})
    return session_params


@pytest.fixture
def socket_name(request: pytest.FixtureRequest) -> str:
    """Random socket name for tmuxp."""
    return f"tmuxp_test{next(namer)}"


# Modules that actually need tmux fixtures in their doctests
DOCTEST_NEEDS_TMUX = {
    "tmuxp.workspace.builder.classic",
}


def pytest_collection_modifyitems(items: list[pytest.Item]) -> None:
    """Add rerun markers to tmux-dependent doctests for flaky shell timing."""
    for item in items:
        if isinstance(item, DoctestItem):
            module_name = item.dtest.globs.get("__name__", "")
            if module_name in DOCTEST_NEEDS_TMUX:
                item.add_marker(pytest.mark.flaky(reruns=2))


@pytest.fixture(autouse=True)
def add_doctest_fixtures(
    request: pytest.FixtureRequest,
    doctest_namespace: dict[str, t.Any],
    tmp_path: pathlib.Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Harness pytest fixtures to doctests namespace."""
    if isinstance(request._pyfuncitem, DoctestItem):
        # Always provide lightweight fixtures
        doctest_namespace["test_utils"] = test_utils
        doctest_namespace["tmp_path"] = tmp_path
        doctest_namespace["monkeypatch"] = monkeypatch

        # Only load expensive tmux fixtures for modules that need them
        module_name = request._pyfuncitem.dtest.globs.get("__name__", "")
        if module_name in DOCTEST_NEEDS_TMUX and shutil.which("tmux"):
            doctest_namespace["server"] = request.getfixturevalue("server")
            session: Session = request.getfixturevalue("session")
            doctest_namespace["session"] = session
            doctest_namespace["window"] = session.active_window
            doctest_namespace["pane"] = session.active_pane



"""CLI utility helpers for tmuxp."""

from __future__ import annotations

import logging
import typing as t

from tmuxp._internal.colors import (
    ColorMode,
    Colors,
    UnknownStyleColor,
    strip_ansi,
    style,
    unstyle,
)
from tmuxp._internal.private_path import PrivatePath
from tmuxp.log import tmuxp_echo

logger = logging.getLogger(__name__)

if t.TYPE_CHECKING:
    from collections.abc import Callable, Sequence

# Re-export for backward compatibility
__all__ = [
    "ColorMode",
    "Colors",
    "UnknownStyleColor",
    "prompt",
    "prompt_bool",
    "prompt_choices",
    "prompt_yes_no",
    "strip_ansi",
    "style",
    "tmuxp_echo",
    "unstyle",
]


def prompt(
    name: str,
    default: str | None = None,
    value_proc: Callable[[str], str] | None = None,
    *,
    color_mode: ColorMode | None = None,
) -> str:
    """Return user input from command line.

    Parameters
    ----------
    name :
        prompt text
    default :
        default value if no input provided.
    color_mode :
        color mode for prompt styling. Defaults to AUTO if not specified.

    Returns
    -------
    str

    See Also
    --------
    prompt_bool
        Prompt for a boolean response.
    prompt_choices
        Prompt for a value from a fixed choice list.
    prompt_yes_no
        Prompt for a yes/no response.

    Notes
    -----
    The prompt helpers are adapted from
    `flask-script <https://github.com/techniq/flask-script>`_; see the
    `flask-script license <https://github.com/techniq/flask-script/blob/master/LICENSE>`_.
    """
    colors = Colors(color_mode if color_mode is not None else ColorMode.AUTO)
    # Use PrivatePath to mask home directory in displayed default
    display_default = str(PrivatePath(default)) if default else None
    prompt_ = name + (
        (display_default and " " + colors.info(f"[{display_default}]")) or ""
    )
    prompt_ += (name.endswith("?") and " ") or ": "
    while True:
        rv = input(prompt_) or default
        # Validate with value_proc only if we have a string value
        if rv is not None and value_proc is not None and callable(value_proc):
            try:
                value_proc(rv)
            except ValueError as e:
                return prompt(
                    str(e),
                    default=default,
                    value_proc=value_proc,
                    color_mode=color_mode,
                )

        if rv:
            return rv
        if default is not None:
            return default
        # No input and no default - loop to re-prompt


def prompt_bool(
    name: str,
    default: bool = False,
    yes_choices: Sequence[t.Any] | None = None,
    no_choices: Sequence[t.Any] | None = None,
    *,
    color_mode: ColorMode | None = None,
) -> bool:
    """Return True / False by prompting user input from command line.

    Parameters
    ----------
    name :
        prompt text
    default :
        default value if no input provided.
    yes_choices :
        default 'y', 'yes', '1', 'on', 'true', 't'
    no_choices :
        default 'n', 'no', '0', 'off', 'false', 'f'
    color_mode :
        color mode for prompt styling. Defaults to AUTO if not specified.

    Returns
    -------
    bool
    """
    colors = Colors(color_mode if color_mode is not None else ColorMode.AUTO)
    yes_choices = yes_choices or ("y", "yes", "1", "on", "true", "t")
    no_choices = no_choices or ("n", "no", "0", "off", "false", "f")

    if default is None:
        prompt_choice = "y/n"
    elif default is True:
        prompt_choice = "Y/n"
    else:
        prompt_choice = "y/N"

    prompt_ = name + " " + colors.muted(f"[{prompt_choice}]")
    prompt_ += (name.endswith("?") and " ") or ": "

    while True:
        rv = input(prompt_)
        if not rv:
            return default
        if rv.lower() in yes_choices:
            return True
        if rv.lower() in no_choices:
            return False


def prompt_yes_no(
    name: str,
    default: bool = True,
    *,
    color_mode: ColorMode | None = None,
) -> bool:
    """:meth:`prompt_bool()` returning yes by default.

    Parameters
    ----------
    name :
        prompt text
    default :
        default value if no input provided.
    color_mode :
        color mode for prompt styling. Defaults to AUTO if not specified.
    """
    return prompt_bool(name, default=default, color_mode=color_mode)


def prompt_choices(
    name: str,
    choices: Sequence[str | tuple[str, str]],
    default: str | None = None,
    no_choice: Sequence[str] = ("none",),
    *,
    color_mode: ColorMode | None = None,
) -> str | None:
    """Return user input from command line from set of provided choices.

    Parameters
    ----------
    name :
        prompt text
    choices :
        Sequence of available choices. Each choice may be a single string or
        a (key, value) tuple where key is used for matching.
    default :
        default value if no input provided.
    no_choice :
        acceptable list of strings for "null choice"
    color_mode :
        color mode for prompt styling. Defaults to AUTO if not specified.

    Returns
    -------
    str
    """
    colors = Colors(color_mode if color_mode is not None else ColorMode.AUTO)
    choices_: list[str] = []
    options: list[str] = []

    for choice in choices:
        if isinstance(choice, str):
            options.append(choice)
        elif isinstance(choice, tuple):
            options.append(f"{choice} [{choice[0]}]")
            choice = choice[0]
        choices_.append(choice)

    choices_str = colors.muted(f"({', '.join(options)})")
    default_str = " " + colors.info(f"[{default}]") if default else ""
    prompt_text = f"{name} - {choices_str}{default_str}"

    while True:
        prompt_ = prompt_text + ": "
        rv = input(prompt_) or default
        if not rv or rv == default:
            return default
        rv = rv.lower()
        if rv in no_choice:
            return None
        if rv in choices_:
            return rv
        tmuxp_echo(
            colors.warning(f"Invalid choice '{rv}'. ")
            + f"Please choose from: {', '.join(choices_)}"
        )



"""Test workspace freezing functionality for tmuxp."""

from __future__ import annotations

import contextlib
import io
import typing as t

import pytest

from tmuxp import cli
from tmuxp._internal.config_reader import ConfigReader

if t.TYPE_CHECKING:
    import pathlib

    from libtmux.server import Server


class FreezeTestFixture(t.NamedTuple):
    """Test fixture for tmuxp freeze command tests."""

    test_id: str
    cli_args: list[str]
    inputs: list[str]


class FreezeOverwriteTestFixture(t.NamedTuple):
    """Test fixture for tmuxp freeze overwrite command tests."""

    test_id: str
    cli_args: list[str]
    inputs: list[str]


FREEZE_TEST_FIXTURES: list[FreezeTestFixture] = [
    FreezeTestFixture(
        test_id="freeze_named_session",
        cli_args=["freeze", "myfrozensession"],
        inputs=["y\n", "./la.yaml\n", "y\n"],
    ),
    FreezeTestFixture(
        test_id="freeze_named_session_exists",
        cli_args=["freeze", "myfrozensession"],
        inputs=["y\n", "./exists.yaml\n", "./la.yaml\n", "y\n"],
    ),
    FreezeTestFixture(
        test_id="freeze_current_session",
        cli_args=["freeze"],
        inputs=["y\n", "./la.yaml\n", "y\n"],
    ),
    FreezeTestFixture(
        test_id="freeze_current_session_exists",
        cli_args=["freeze"],
        inputs=["y\n", "./exists.yaml\n", "./la.yaml\n", "y\n"],
    ),
]


FREEZE_OVERWRITE_TEST_FIXTURES: list[FreezeOverwriteTestFixture] = [
    FreezeOverwriteTestFixture(
        test_id="force_overwrite_named_session",
        cli_args=["freeze", "mysession", "--force"],
        inputs=["\n", "\n", "y\n", "./exists.yaml\n", "y\n"],
    ),
    FreezeOverwriteTestFixture(
        test_id="force_overwrite_current_session",
        cli_args=["freeze", "--force"],
        inputs=["\n", "\n", "y\n", "./exists.yaml\n", "y\n"],
    ),
]


@pytest.mark.parametrize(
    list(FreezeTestFixture._fields),
    FREEZE_TEST_FIXTURES,
    ids=[test.test_id for test in FREEZE_TEST_FIXTURES],
)
def test_freeze(
    server: Server,
    test_id: str,
    cli_args: list[str],
    inputs: list[str],
    tmp_path: pathlib.Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Parametrized test for freezing a tmux session to a tmuxp config file."""
    monkeypatch.setenv("HOME", str(tmp_path))
    exists_yaml = tmp_path / "exists.yaml"
    exists_yaml.touch()

    server.new_session(session_name="myfirstsession")
    server.new_session(session_name="myfrozensession")

    # Assign an active pane to the session
    second_session = server.sessions[1]
    first_pane_on_second_session_id = second_session.windows[0].panes[0].pane_id
    assert first_pane_on_second_session_id
    monkeypatch.setenv("TMUX_PANE", first_pane_on_second_session_id)

    monkeypatch.chdir(tmp_path)
    # Use tmux server (socket name) used in the test
    assert server.socket_name is not None
    cli_args = [*cli_args, "-L", server.socket_name]

    monkeypatch.setattr("sys.stdin", io.StringIO("".join(inputs)))
    with contextlib.suppress(SystemExit):
        cli.cli(cli_args)

    yaml_config_path = tmp_path / "la.yaml"
    assert yaml_config_path.exists()

    yaml_config = yaml_config_path.open().read()
    frozen_config = ConfigReader._load(fmt="yaml", content=yaml_config)

    assert frozen_config["session_name"] == "myfrozensession"


@pytest.mark.parametrize(
    list(FreezeOverwriteTestFixture._fields),
    FREEZE_OVERWRITE_TEST_FIXTURES,
    ids=[test.test_id for test in FREEZE_OVERWRITE_TEST_FIXTURES],
)
def test_freeze_overwrite(
    server: Server,
    test_id: str,
    cli_args: list[str],
    inputs: list[str],
    tmp_path: pathlib.Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Test overwrite prompt when freezing a tmuxp configuration file."""
    monkeypatch.setenv("HOME", str(tmp_path))
    exists_yaml = tmp_path / "exists.yaml"
    exists_yaml.touch()

    server.new_session(session_name="mysession")

    monkeypatch.chdir(tmp_path)
    # Use tmux server (socket name) used in the test
    assert server.socket_name is not None
    cli_args = [*cli_args, "-L", server.socket_name]

    monkeypatch.setattr("sys.stdin", io.StringIO("".join(inputs)))
    with contextlib.suppress(SystemExit):
        cli.cli(cli_args)

    yaml_config_path = tmp_path / "exists.yaml"
    assert yaml_config_path.exists()



"""CLI tests for tmuxp search command."""

from __future__ import annotations

import json
import pathlib
import re
import typing as t

import pytest

from tmuxp.cli._colors import ColorMode, Colors
from tmuxp.cli._output import OutputFormatter, OutputMode
from tmuxp.cli.search import (
    DEFAULT_FIELDS,
    InvalidFieldError,
    SearchPattern,
    SearchToken,
    WorkspaceFields,
    WorkspaceSearchResult,
    _get_field_values,
    _output_search_results,
    compile_search_patterns,
    create_search_subparser,
    evaluate_match,
    extract_workspace_fields,
    find_search_matches,
    highlight_matches,
    normalize_fields,
    parse_query_terms,
)


class NormalizeFieldsFixture(t.NamedTuple):
    """Test fixture for normalize_fields."""

    test_id: str
    fields: list[str] | None
    expected: tuple[str, ...]
    raises: type[Exception] | None


NORMALIZE_FIELDS_FIXTURES: list[NormalizeFieldsFixture] = [
    NormalizeFieldsFixture(
        test_id="none_returns_defaults",
        fields=None,
        expected=DEFAULT_FIELDS,
        raises=None,
    ),
    NormalizeFieldsFixture(
        test_id="name_alias",
        fields=["n"],
        expected=("name",),
        raises=None,
    ),
    NormalizeFieldsFixture(
        test_id="session_aliases",
        fields=["s", "session", "session_name"],
        expected=("session_name",),
        raises=None,
    ),
    NormalizeFieldsFixture(
        test_id="path_alias",
        fields=["p"],
        expected=("path",),
        raises=None,
    ),
    NormalizeFieldsFixture(
        test_id="window_alias",
        fields=["w"],
        expected=("window",),
        raises=None,
    ),
    NormalizeFieldsFixture(
        test_id="multiple_fields",
        fields=["name", "s", "window"],
        expected=("name", "session_name", "window"),
        raises=None,
    ),
    NormalizeFieldsFixture(
        test_id="invalid_field",
        fields=["invalid"],
        expected=(),
        raises=InvalidFieldError,
    ),
    NormalizeFieldsFixture(
        test_id="case_insensitive",
        fields=["NAME", "Session"],
        expected=("name", "session_name"),
        raises=None,
    ),
]


@pytest.mark.parametrize(
    NormalizeFieldsFixture._fields,
    NORMALIZE_FIELDS_FIXTURES,
    ids=[test.test_id for test in NORMALIZE_FIELDS_FIXTURES],
)
def test_normalize_fields(
    test_id: str,
    fields: list[str] | None,
    expected: tuple[str, ...],
    raises: type[Exception] | None,
) -> None:
    """Test normalize_fields function."""
    if raises:
        with pytest.raises(raises):
            normalize_fields(fields)
    else:
        result = normalize_fields(fields)
        assert result == expected


class ParseQueryTermsFixture(t.NamedTuple):
    """Test fixture for parse_query_terms."""

    test_id: str
    terms: list[str]
    expected_count: int
    expected_first_fields: tuple[str, ...] | None
    expected_first_pattern: str | None


PARSE_QUERY_TERMS_FIXTURES: list[ParseQueryTermsFixture] = [
    ParseQueryTermsFixture(
        test_id="simple_term",
        terms=["dev"],
        expected_count=1,
        expected_first_fields=DEFAULT_FIELDS,
        expected_first_pattern="dev",
    ),
    ParseQueryTermsFixture(
        test_id="name_prefix",
        terms=["name:dev"],
        expected_count=1,
        expected_first_fields=("name",),
        expected_first_pattern="dev",
    ),
    ParseQueryTermsFixture(
        test_id="session_prefix",
        terms=["s:production"],
        expected_count=1,
        expected_first_fields=("session_name",),
        expected_first_pattern="production",
    ),
    ParseQueryTermsFixture(
        test_id="multiple_terms",
        terms=["dev", "production"],
        expected_count=2,
        expected_first_fields=DEFAULT_FIELDS,
        expected_first_pattern="dev",
    ),
    ParseQueryTermsFixture(
        test_id="url_not_field",
        terms=["http://example.com"],
        expected_count=1,
        expected_first_fields=DEFAULT_FIELDS,
        expected_first_pattern="http://example.com",
    ),
    ParseQueryTermsFixture(
        test_id="empty_pattern_skipped",
        terms=["name:"],
        expected_count=0,
        expected_first_fields=None,
        expected_first_pattern=None,
    ),
    ParseQueryTermsFixture(
        test_id="path_with_colons",
        terms=["path:/home/user/project"],
        expected_count=1,
        expected_first_fields=("path",),
        expected_first_pattern="/home/user/project",
    ),
]


@pytest.mark.parametrize(
    ParseQueryTermsFixture._fields,
    PARSE_QUERY_TERMS_FIXTURES,
    ids=[test.test_id for test in PARSE_QUERY_TERMS_FIXTURES],
)
def test_parse_query_terms(
    test_id: str,
    terms: list[str],
    expected_count: int,
    expected_first_fields: tuple[str, ...] | None,
    expected_first_pattern: str | None,
) -> None:
    """Test parse_query_terms function."""
    result = parse_query_terms(terms)

    assert len(result) == expected_count

    if expected_count > 0:
        assert result[0].fields == expected_first_fields
        assert result[0].pattern == expected_first_pattern


class CompileSearchPatternsFixture(t.NamedTuple):
    """Test fixture for compile_search_patterns."""

    test_id: str
    pattern: str
    ignore_case: bool
    smart_case: bool
    fixed_strings: bool
    word_regexp: bool
    test_string: str
    should_match: bool


COMPILE_SEARCH_PATTERNS_FIXTURES: list[CompileSearchPatternsFixture] = [
    CompileSearchPatternsFixture(
        test_id="basic_match",
        pattern="dev",
        ignore_case=False,
        smart_case=False,
        fixed_strings=False,
        word_regexp=False,
        test_string="development",
        should_match=True,
    ),
    CompileSearchPatternsFixture(
        test_id="case_sensitive_no_match",
        pattern="DEV",
        ignore_case=False,
        smart_case=False,
        fixed_strings=False,
        word_regexp=False,
        test_string="development",
        should_match=False,
    ),
    CompileSearchPatternsFixture(
        test_id="ignore_case_match",
        pattern="DEV",
        ignore_case=True,
        smart_case=False,
        fixed_strings=False,
        word_regexp=False,
        test_string="development",
        should_match=True,
    ),
    CompileSearchPatternsFixture(
        test_id="smart_case_lowercase",
        pattern="dev",
        ignore_case=False,
        smart_case=True,
        fixed_strings=False,
        word_regexp=False,
        test_string="DEVELOPMENT",
        should_match=True,
    ),
    CompileSearchPatternsFixture(
        test_id="smart_case_uppercase_no_match",
        pattern="Dev",
        ignore_case=False,
        smart_case=True,
        fixed_strings=False,
        word_regexp=False,
        test_string="development",
        should_match=False,
    ),
    CompileSearchPatternsFixture(
        test_id="fixed_strings_literal",
        pattern="dev.*",
        ignore_case=False,
        smart_case=False,
        fixed_strings=True,
        word_regexp=False,
        test_string="dev.*project",
        should_match=True,
    ),
    CompileSearchPatternsFixture(
        test_id="fixed_strings_no_regex",
        pattern="dev.*",
        ignore_case=False,
        smart_case=False,
        fixed_strings=True,
        word_regexp=False,
        test_string="development",
        should_match=False,
    ),
    CompileSearchPatternsFixture(
        test_id="word_boundary_match",
        pattern="dev",
        ignore_case=False,
        smart_case=False,
        fixed_strings=False,
        word_regexp=True,
        test_string="my dev project",
        should_match=True,
    ),
    CompileSearchPatternsFixture(
        test_id="word_boundary_no_match",
        pattern="dev",
        ignore_case=False,
        smart_case=False,
        fixed_strings=False,
        word_regexp=True,
        test_string="development",
        should_match=False,
    ),
    CompileSearchPatternsFixture(
        test_id="regex_pattern",
        pattern="dev.*proj",
        ignore_case=False,
        smart_case=False,
        fixed_strings=False,
        word_regexp=False,
        test_string="dev-project",
        should_match=True,
    ),
]


@pytest.mark.parametrize(
    CompileSearchPatternsFixture._fields,
    COMPILE_SEARCH_PATTERNS_FIXTURES,
    ids=[test.test_id for test in COMPILE_SEARCH_PATTERNS_FIXTURES],
)
def test_compile_search_patterns(
    test_id: str,
    pattern: str,
    ignore_case: bool,
    smart_case: bool,
    fixed_strings: bool,
    word_regexp: bool,
    test_string: str,
    should_match: bool,
) -> None:
    """Test compile_search_patterns function."""
    tokens = [SearchToken(fields=("name",), pattern=pattern)]

    patterns = compile_search_patterns(
        tokens,
        ignore_case=ignore_case,
        smart_case=smart_case,
        fixed_strings=fixed_strings,
        word_regexp=word_regexp,
    )

    assert len(patterns) == 1
    match = patterns[0].regex.search(test_string)
    assert bool(match) == should_match


def test_compile_search_patterns_invalid_regex_raises() -> None:
    """Invalid regex pattern raises re.error."""
    tokens = [SearchToken(fields=("name",), pattern="[invalid(")]
    with pytest.raises(re.error):
        compile_search_patterns(tokens)


def test_extract_workspace_fields_basic(tmp_path: pathlib.Path) -> None:
    """Extract fields from basic workspace file."""
    workspace = tmp_path / "test.yaml"
    workspace.write_text(
        "session_name: my-session\n"
        "windows:\n"
        "  - window_name: editor\n"
        "    panes:\n"
        "      - vim\n"
        "  - window_name: shell\n"
    )

    fields = extract_workspace_fields(workspace)

    assert fields["name"] == "test"
    assert fields["session_name"] == "my-session"
    assert "editor" in fields["windows"]
    assert "shell" in fields["windows"]
    assert "vim" in fields["panes"]


def test_extract_workspace_fields_pane_shell_command_dict(
    tmp_path: pathlib.Path,
) -> None:
    """Extract pane commands from dict format."""
    workspace = tmp_path / "test.yaml"
    workspace.write_text(
        "session_name: test\n"
        "windows:\n"
        "  - window_name: main\n"
        "    panes:\n"
        "      - shell_command: git status\n"
        "      - shell_command:\n"
        "          - npm install\n"
        "          - npm start\n"
    )

    fields = extract_workspace_fields(workspace)

    assert "git status" in fields["panes"]
    assert "npm install" in fields["panes"]
    assert "npm start" in fields["panes"]


def test_extract_workspace_fields_missing_session_name(tmp_path: pathlib.Path) -> None:
    """Handle workspace without session_name."""
    workspace = tmp_path / "test.yaml"
    workspace.write_text("windows:\n  - window_name: main\n")

    fields = extract_workspace_fields(workspace)

    assert fields["session_name"] == ""
    assert fields["name"] == "test"


def test_extract_workspace_fields_invalid_yaml(tmp_path: pathlib.Path) -> None:
    """Handle invalid YAML gracefully."""
    workspace = tmp_path / "test.yaml"
    workspace.write_text("{{{{invalid yaml")

    fields = extract_workspace_fields(workspace)

    assert fields["name"] == "test"
    assert fields["session_name"] == ""
    assert fields["windows"] == []


def test_extract_workspace_fields_path_uses_privacy(
    tmp_path: pathlib.Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Path should use PrivatePath for home contraction."""
    monkeypatch.setattr(pathlib.Path, "home", lambda: tmp_path)
    workspace = tmp_path / "test.yaml"
    workspace.write_text("session_name: test\n")

    fields = extract_workspace_fields(workspace)

    assert fields["path"] == "~/test.yaml"


@pytest.fixture()
def sample_fields() -> WorkspaceFields:
    """Sample workspace fields for testing."""
    return WorkspaceFields(
        name="dev-project",
        path="~/.tmuxp/dev-project.yaml",
        session_name="development",
        windows=["editor", "shell", "logs"],
        panes=["vim", "git status", "tail -f"],
    )


def test_evaluate_match_single_pattern(sample_fields: WorkspaceFields) -> None:
    """Single pattern should match."""
    pattern = SearchPattern(
        fields=("name",),
        raw="dev",
        regex=re.compile("dev"),
    )

    matched, matches = evaluate_match(sample_fields, [pattern])

    assert matched is True
    assert "name" in matches


def test_evaluate_match_single_pattern_no_match(sample_fields: WorkspaceFields) -> None:
    """Single pattern should not match."""
    pattern = SearchPattern(
        fields=("name",),
        raw="xyz",
        regex=re.compile("xyz"),
    )

    matched, matches = evaluate_match(sample_fields, [pattern])

    assert matched is False
    assert matches == {}


def test_evaluate_match_and_logic_all_match(sample_fields: WorkspaceFields) -> None:
    """AND logic - all patterns match."""
    p1 = SearchPattern(fields=("name",), raw="dev", regex=re.compile("dev"))
    p2 = SearchPattern(fields=("name",), raw="project", regex=re.compile("project"))

    matched, _ = evaluate_match(sample_fields, [p1, p2], match_any=False)

    assert matched is True


def test_evaluate_match_and_logic_partial_no_match(
    sample_fields: WorkspaceFields,
) -> None:
    """AND logic - only some patterns match."""
    p1 = SearchPattern(fields=("name",), raw="dev", regex=re.compile("dev"))
    p2 = SearchPattern(fields=("name",), raw="xyz", regex=re.compile("xyz"))

    matched, _ = evaluate_match(sample_fields, [p1, p2], match_any=False)

    assert matched is False


def test_evaluate_match_or_logic_any_match(sample_fields: WorkspaceFields) -> None:
    """OR logic - any pattern matches."""
    p1 = SearchPattern(fields=("name",), raw="xyz", regex=re.compile("xyz"))
    p2 = SearchPattern(fields=("name",), raw="dev", regex=re.compile("dev"))

    matched, _ = evaluate_match(sample_fields, [p1, p2], match_any=True)

    assert matched is True


def test_evaluate_match_window_field(sample_fields: WorkspaceFields) -> None:
    """Search in window field."""
    pattern = SearchPattern(
        fields=("window",),
        raw="editor",
        regex=re.compile("editor"),
    )

    matched, matches = evaluate_match(sample_fields, [pattern])

    assert matched is True
    assert "window" in matches


def test_evaluate_match_pane_field(sample_fields: WorkspaceFields) -> None:
    """Search in pane field."""
    pattern = SearchPattern(
        fields=("pane",),
        raw="vim",
        regex=re.compile("vim"),
    )

    matched, matches = evaluate_match(sample_fields, [pattern])

    assert matched is True
    assert "pane" in matches


def test_evaluate_match_multiple_fields(sample_fields: WorkspaceFields) -> None:
    """Pattern searches multiple fields."""
    pattern = SearchPattern(
        fields=("name", "session_name"),
        raw="dev",
        regex=re.compile("dev"),
    )

    matched, matches = evaluate_match(sample_fields, [pattern])

    assert matched is True
    # Should find matches in both name and session_name
    assert "name" in matches or "session_name" in matches


def test_find_search_matches_basic(tmp_path: pathlib.Path) -> None:
    """Basic search finds matching workspace."""
    workspace = tmp_path / "dev.yaml"
    workspace.write_text("session_name: development\n")

    pattern = SearchPattern(
        fields=("session_name",),
        raw="dev",
        regex=re.compile("dev"),
    )

    results = find_search_matches([(workspace, "global")], [pattern])

    assert len(results) == 1
    assert results[0]["source"] == "global"


def test_find_search_matches_no_match(tmp_path: pathlib.Path) -> None:
    """Search returns empty when no match."""
    workspace = tmp_path / "production.yaml"
    workspace.write_text("session_name: production\n")

    pattern = SearchPattern(
        fields=("name",),
        raw="dev",
        regex=re.compile("dev"),
    )

    results = find_search_matches([(workspace, "global")], [pattern])

    assert len(results) == 0


def test_find_search_matches_invert(tmp_path: pathlib.Path) -> None:
    """Invert match returns non-matching workspaces."""
    workspace = tmp_path / "production.yaml"
    workspace.write_text("session_name: production\n")

    pattern = SearchPattern(
        fields=("name",),
        raw="dev",
        regex=re.compile("dev"),
    )

    results = find_search_matches([(workspace, "global")], [pattern], invert_match=True)

    assert len(results) == 1


def test_find_search_matches_multiple_workspaces(tmp_path: pathlib.Path) -> None:
    """Search across multiple workspaces."""
    ws1 = tmp_path / "dev.yaml"
    ws1.write_text("session_name: development\n")

    ws2 = tmp_path / "prod.yaml"
    ws2.write_text("session_name: production\n")

    pattern = SearchPattern(
        fields=("name", "session_name"),
        raw="dev",
        regex=re.compile("dev"),
    )

    results = find_search_matches([(ws1, "global"), (ws2, "global")], [pattern])

    assert len(results) == 1
    assert results[0]["fields"]["name"] == "dev"


def test_highlight_matches_no_colors() -> None:
    """Colors disabled returns original text."""
    colors = Colors(ColorMode.NEVER)
    pattern = SearchPattern(
        fields=("name",),
        raw="dev",
        regex=re.compile("dev"),
    )

    result = highlight_matches("development", [pattern], colors=colors)

    assert result == "development"


def test_highlight_matches_with_colors() -> None:
    """Colors enabled adds ANSI codes."""
    colors = Colors(ColorMode.ALWAYS)
    pattern = SearchPattern(
        fields=("name",),
        raw="dev",
        regex=re.compile("dev"),
    )

    result = highlight_matches("development", [pattern], colors=colors)

    assert "\033[" in result  # Contains ANSI escape
    assert "dev" in result


def test_highlight_matches_no_match() -> None:
    """No match returns original text."""
    colors = Colors(ColorMode.ALWAYS)
    pattern = SearchPattern(
        fields=("name",),
        raw="xyz",
        regex=re.compile("xyz"),
    )

    result = highlight_matches("development", [pattern], colors=colors)

    assert result == "development"


def test_highlight_matches_multiple() -> None:
    """Multiple matches in same string."""
    colors = Colors(ColorMode.ALWAYS)
    pattern = SearchPattern(
        fields=("name",),
        raw="e",
        regex=re.compile("e"),
    )

    result = highlight_matches("development", [pattern], colors=colors)

    # Should contain multiple highlights
    assert result.count("\033[") > 1


def test_highlight_matches_empty_patterns() -> None:
    """Empty patterns returns original text."""
    colors = Colors(ColorMode.ALWAYS)

    result = highlight_matches("development", [], colors=colors)

    assert result == "development"


@pytest.fixture()
def sample_fields_for_get_field_values() -> WorkspaceFields:
    """Sample workspace fields."""
    return WorkspaceFields(
        name="test",
        path="~/.tmuxp/test.yaml",
        session_name="test-session",
        windows=["editor", "shell"],
        panes=["vim", "bash"],
    )


def test_get_field_values_scalar(
    sample_fields_for_get_field_values: WorkspaceFields,
) -> None:
    """Scalar field returns list with one item."""
    result = _get_field_values(sample_fields_for_get_field_values, "name")
    assert result == ["test"]


def test_get_field_values_list(
    sample_fields_for_get_field_values: WorkspaceFields,
) -> None:
    """List field returns the list."""
    result = _get_field_values(sample_fields_for_get_field_values, "windows")
    assert result == ["editor", "shell"]


def test_get_field_values_window_alias(
    sample_fields_for_get_field_values: WorkspaceFields,
) -> None:
    """Window alias maps to windows."""
    result = _get_field_values(sample_fields_for_get_field_values, "window")
    assert result == ["editor", "shell"]


def test_get_field_values_pane_alias(
    sample_fields_for_get_field_values: WorkspaceFields,
) -> None:
    """Pane alias maps to panes."""
    result = _get_field_values(sample_fields_for_get_field_values, "pane")
    assert result == ["vim", "bash"]


def test_get_field_values_empty() -> None:
    """Empty value returns empty list."""
    fields = WorkspaceFields(
        name="",
        path="",
        session_name="",
        windows=[],
        panes=[],
    )
    result = _get_field_values(fields, "name")
    assert result == []


def test_search_subparser_creation() -> None:
    """Subparser can be created successfully."""
    import argparse

    parser = argparse.ArgumentParser()
    result = create_search_subparser(parser)

    assert result is parser


def test_search_subparser_options() -> None:
    """Parser has expected options."""
    import argparse

    parser = argparse.ArgumentParser()
    create_search_subparser(parser)

    # Parse with various options
    args = parser.parse_args(["-i", "-S", "-F", "-w", "-v", "--any", "pattern"])

    assert args.ignore_case is True
    assert args.smart_case is True
    assert args.fixed_strings is True
    assert args.word_regexp is True
    assert args.invert_match is True
    assert args.match_any is True
    assert args.query_terms == ["pattern"]


def test_search_subparser_output_format_options() -> None:
    """Parser supports output format options."""
    import argparse

    parser = argparse.ArgumentParser()
    create_search_subparser(parser)

    args_json = parser.parse_args(["--json", "test"])
    assert args_json.output_json is True

    args_ndjson = parser.parse_args(["--ndjson", "test"])
    assert args_ndjson.output_ndjson is True


def test_search_subparser_field_option() -> None:
    """Parser supports field option."""
    import argparse

    parser = argparse.ArgumentParser()
    create_search_subparser(parser)

    args = parser.parse_args(["-f", "name", "-f", "session", "test"])

    assert args.field == ["name", "session"]


def test_output_search_results_no_results(capsys: pytest.CaptureFixture[str]) -> None:
    """No results outputs warning message."""
    colors = Colors(ColorMode.NEVER)
    formatter = OutputFormatter(OutputMode.HUMAN)

    _output_search_results([], [], formatter, colors)
    formatter.finalize()

    captured = capsys.readouterr()
    assert "No matching" in captured.out


def test_output_search_results_json(capsys: pytest.CaptureFixture[str]) -> None:
    """JSON output mode produces valid JSON."""
    colors = Colors(ColorMode.NEVER)
    formatter = OutputFormatter(OutputMode.JSON)

    result: WorkspaceSearchResult = {
        "filepath": "/test/dev.yaml",
        "source": "global",
        "fields": WorkspaceFields(
            name="dev",
            path="~/.tmuxp/dev.yaml",
            session_name="development",
            windows=["editor"],
            panes=["vim"],
        ),
        "matches": {"name": ["dev"]},
    }

    _output_search_results([result], [], formatter, colors)
    formatter.finalize()

    captured = capsys.readouterr()
    assert "\x1b" not in captured.out, "ANSI escapes must not leak into machine output"
    data = json.loads(captured.out)
    assert len(data) == 1
    assert data[0]["name"] == "dev"


def test_output_search_results_ndjson(capsys: pytest.CaptureFixture[str]) -> None:
    """NDJSON output mode produces one JSON per line."""
    colors = Colors(ColorMode.NEVER)
    formatter = OutputFormatter(OutputMode.NDJSON)

    result: WorkspaceSearchResult = {
        "filepath": "/test/dev.yaml",
        "source": "global",
        "fields": WorkspaceFields(
            name="dev",
            path="~/.tmuxp/dev.yaml",
            session_name="development",
            windows=[],
            panes=[],
        ),
        "matches": {"name": ["dev"]},
    }

    _output_search_results([result], [], formatter, colors)
    formatter.finalize()

    captured = capsys.readouterr()
    assert "\x1b" not in captured.out, "ANSI escapes must not leak into machine output"
    lines = captured.out.strip().split("\n")
    # Filter out human-readable lines
    json_lines = [line for line in lines if line.startswith("{")]
    assert len(json_lines) >= 1
    data = json.loads(json_lines[0])
    assert data["name"] == "dev"



"""Tests for PrivatePath privacy-masking utilities."""

from __future__ import annotations

import pathlib

import pytest

from tmuxp._internal.private_path import PrivatePath, collapse_home_in_string

# PrivatePath tests


def test_private_path_collapses_home(monkeypatch: pytest.MonkeyPatch) -> None:
    """PrivatePath replaces home directory with ~."""
    monkeypatch.setattr(pathlib.Path, "home", lambda: pathlib.Path("/home/testuser"))

    path = PrivatePath("/home/testuser/projects/tmuxp")
    assert str(path) == "~/projects/tmuxp"


def test_private_path_collapses_home_exact(monkeypatch: pytest.MonkeyPatch) -> None:
    """PrivatePath handles exact home directory match."""
    monkeypatch.setattr(pathlib.Path, "home", lambda: pathlib.Path("/home/testuser"))

    path = PrivatePath("/home/testuser")
    assert str(path) == "~"


def test_private_path_preserves_non_home(monkeypatch: pytest.MonkeyPatch) -> None:
    """PrivatePath preserves paths outside home directory."""
    monkeypatch.setattr(pathlib.Path, "home", lambda: pathlib.Path("/home/testuser"))

    path = PrivatePath("/usr/bin/tmux")
    assert str(path) == "/usr/bin/tmux"


def test_private_path_preserves_tmp(monkeypatch: pytest.MonkeyPatch) -> None:
    """PrivatePath preserves /tmp paths."""
    monkeypatch.setattr(pathlib.Path, "home", lambda: pathlib.Path("/home/testuser"))

    path = PrivatePath("/tmp/example")
    assert str(path) == "/tmp/example"


def test_private_path_preserves_already_collapsed() -> None:
    """PrivatePath preserves paths already starting with ~."""
    path = PrivatePath("~/already/collapsed")
    assert str(path) == "~/already/collapsed"


def test_private_path_repr(monkeypatch: pytest.MonkeyPatch) -> None:
    """PrivatePath repr shows class name and collapsed path."""
    monkeypatch.setattr(pathlib.Path, "home", lambda: pathlib.Path("/home/testuser"))

    path = PrivatePath("/home/testuser/config.yaml")
    assert repr(path) == "PrivatePath('~/config.yaml')"


def test_private_path_in_fstring(monkeypatch: pytest.MonkeyPatch) -> None:
    """PrivatePath works in f-strings with collapsed home."""
    monkeypatch.setattr(pathlib.Path, "home", lambda: pathlib.Path("/home/testuser"))

    path = PrivatePath("/home/testuser/.tmuxp/session.yaml")
    result = f"config: {path}"
    assert result == "config: ~/.tmuxp/session.yaml"


def test_private_path_similar_prefix_not_collapsed(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """PrivatePath does not collapse paths with similar prefix but different user."""
    monkeypatch.setattr(pathlib.Path, "home", lambda: pathlib.Path("/home/testuser"))

    # /home/testuser2 should NOT be collapsed even though it starts with /home/testuser
    path = PrivatePath("/home/testuser2/projects")
    assert str(path) == "/home/testuser2/projects"


# collapse_home_in_string tests


def test_collapse_home_in_string_single_path(monkeypatch: pytest.MonkeyPatch) -> None:
    """collapse_home_in_string handles a single path."""
    monkeypatch.setattr(pathlib.Path, "home", lambda: pathlib.Path("/home/testuser"))

    result = collapse_home_in_string("/home/testuser/.local/bin")
    assert result == "~/.local/bin"


def test_collapse_home_in_string_multiple_paths(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """collapse_home_in_string handles colon-separated paths."""
    monkeypatch.setattr(pathlib.Path, "home", lambda: pathlib.Path("/home/testuser"))

    result = collapse_home_in_string(
        "/home/testuser/bin:/home/testuser/.cargo/bin:/usr/bin"
    )
    assert result == "~/bin:~/.cargo/bin:/usr/bin"


def test_collapse_home_in_string_no_home_paths(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """collapse_home_in_string preserves paths not under home."""
    monkeypatch.setattr(pathlib.Path, "home", lambda: pathlib.Path("/home/testuser"))

    result = collapse_home_in_string("/usr/bin:/bin:/usr/local/bin")
    assert result == "/usr/bin:/bin:/usr/local/bin"


def test_collapse_home_in_string_mixed_paths(monkeypatch: pytest.MonkeyPatch) -> None:
    """collapse_home_in_string handles mixed home and non-home paths."""
    monkeypatch.setattr(pathlib.Path, "home", lambda: pathlib.Path("/home/testuser"))

    result = collapse_home_in_string("/usr/bin:/home/testuser/.local/bin:/bin")
    assert result == "/usr/bin:~/.local/bin:/bin"


def test_collapse_home_in_string_empty() -> None:
    """collapse_home_in_string handles empty string."""
    result = collapse_home_in_string("")
    assert result == ""



"""Tests for CLI colors in freeze command."""

from __future__ import annotations

import pathlib

from tests.cli.conftest import (
    ANSI_BLUE,
    ANSI_CYAN,
    ANSI_GREEN,
    ANSI_RED,
    ANSI_RESET,
    ANSI_YELLOW,
)
from tmuxp._internal.private_path import PrivatePath
from tmuxp.cli._colors import Colors

# Freeze command color output tests


def test_freeze_error_uses_red(colors_always: Colors) -> None:
    """Verify error messages use error color (red)."""
    msg = "Session not found"
    result = colors_always.error(msg)
    assert ANSI_RED in result  # red foreground
    assert msg in result
    assert result.endswith(ANSI_RESET)  # reset at end


def test_freeze_success_message(colors_always: Colors) -> None:
    """Verify success messages use success color (green)."""
    result = colors_always.success("Saved to ")
    assert ANSI_GREEN in result  # green foreground
    assert "Saved to" in result


def test_freeze_file_path_uses_info(colors_always: Colors) -> None:
    """Verify file paths use info color (cyan)."""
    path = "/path/to/config.yaml"
    result = colors_always.info(path)
    assert ANSI_CYAN in result  # cyan foreground
    assert path in result


def test_freeze_warning_file_exists(colors_always: Colors) -> None:
    """Verify file exists warning uses warning color (yellow)."""
    msg = "/path/to/config.yaml exists."
    result = colors_always.warning(msg)
    assert ANSI_YELLOW in result  # yellow foreground
    assert msg in result


def test_freeze_muted_for_secondary_text(colors_always: Colors) -> None:
    """Verify secondary text uses muted color (blue)."""
    msg = "Freeze does its best to snapshot live tmux sessions."
    result = colors_always.muted(msg)
    assert ANSI_BLUE in result  # blue foreground
    assert msg in result


def test_freeze_colors_disabled_plain_text(colors_never: Colors) -> None:
    """Verify disabled colors return plain text."""
    assert colors_never.error("error") == "error"
    assert colors_never.success("success") == "success"
    assert colors_never.warning("warning") == "warning"
    assert colors_never.info("info") == "info"
    assert colors_never.muted("muted") == "muted"


def test_freeze_combined_output_format(colors_always: Colors) -> None:
    """Verify combined success + info format for 'Saved to <path>' message."""
    dest = "/home/user/.tmuxp/session.yaml"
    output = colors_always.success("Saved to ") + colors_always.info(dest) + "."
    # Should contain both green and cyan ANSI codes
    assert ANSI_GREEN in output  # green for "Saved to"
    assert ANSI_CYAN in output  # cyan for path
    assert "Saved to" in output
    assert dest in output
    assert output.endswith(".")


def test_freeze_warning_with_instructions(colors_always: Colors) -> None:
    """Verify warning + muted format for file exists message."""
    path = "/path/to/config.yaml"
    output = (
        colors_always.warning(f"{path} exists.")
        + " "
        + colors_always.muted("Pick a new filename.")
    )
    # Should contain both yellow and blue ANSI codes
    assert ANSI_YELLOW in output  # yellow for warning
    assert ANSI_BLUE in output  # blue for muted
    assert path in output
    assert "Pick a new filename." in output


def test_freeze_url_highlighted_in_help(colors_always: Colors) -> None:
    """Verify URLs use info color in help text."""
    url = "<http://tmuxp.git-pull.com/examples.html>"
    help_text = colors_always.muted("tmuxp has examples at ") + colors_always.info(url)
    assert ANSI_BLUE in help_text  # blue for muted text
    assert ANSI_CYAN in help_text  # cyan for URL
    assert url in help_text


# Privacy masking tests


def test_freeze_masks_home_in_saved_message(
    colors_always: Colors,
    mock_home: pathlib.Path,
) -> None:
    """Freeze should mask home directory in 'Saved to' message."""
    dest = mock_home / ".tmuxp/session.yaml"
    output = (
        colors_always.success("Saved to ")
        + colors_always.info(str(PrivatePath(dest)))
        + "."
    )

    assert "~/.tmuxp/session.yaml" in output
    assert "/home/testuser" not in output


def test_freeze_masks_home_in_exists_warning(
    colors_always: Colors,
    mock_home: pathlib.Path,
) -> None:
    """Freeze should mask home directory in 'exists' warning."""
    dest_prompt = mock_home / ".tmuxp/session.yaml"
    output = colors_always.warning(f"{PrivatePath(dest_prompt)} exists.")

    assert "~/.tmuxp/session.yaml exists." in output
    assert "/home/testuser" not in output


def test_freeze_masks_home_in_save_to_prompt(mock_home: pathlib.Path) -> None:
    """Freeze should mask home directory in 'Save to:' prompt."""
    save_to = mock_home / ".tmuxp/session.yaml"
    prompt_text = f"Save to: {PrivatePath(save_to)}"

    assert "~/.tmuxp/session.yaml" in prompt_text
    assert "/home/testuser" not in prompt_text


def test_freeze_masks_home_in_save_confirmation(mock_home: pathlib.Path) -> None:
    """Freeze should mask home directory in 'Save to ...?' confirmation."""
    dest = mock_home / ".tmuxp/session.yaml"
    prompt_text = f"Save to {PrivatePath(dest)}?"

    assert "~/.tmuxp/session.yaml" in prompt_text
    assert "/home/testuser" not in prompt_text



"""Workspace builders: turn an expanded workspace ``dict`` into a tmux session.

:class:`~tmuxp.workspace.builder.classic.ClassicWorkspaceBuilder` is the default
implementation. A workspace may select a different builder with the
``workspace_builder`` config key, resolved by
:mod:`tmuxp.workspace.builder.registry`.

``WorkspaceBuilder`` remains importable here as a backwards-compatible alias of
:class:`~tmuxp.workspace.builder.classic.ClassicWorkspaceBuilder`.
"""

from __future__ import annotations

from tmuxp.workspace.builder.classic import (
    ClassicWorkspaceBuilder,
    get_default_columns,
    get_default_rows,
)
from tmuxp.workspace.builder.protocol import WorkspaceBuilderProtocol
from tmuxp.workspace.builder.registry import (
    WORKSPACE_BUILDERS_GROUP,
    available_builders,
    prepended_sys_path,
    resolve_builder_class,
    resolve_builder_paths,
)

# Backwards-compatible alias: the classic builder was historically named
# ``WorkspaceBuilder`` and imported from ``tmuxp.workspace.builder``.
WorkspaceBuilder = ClassicWorkspaceBuilder

__all__ = [
    "WORKSPACE_BUILDERS_GROUP",
    "ClassicWorkspaceBuilder",
    "WorkspaceBuilder",
    "WorkspaceBuilderProtocol",
    "available_builders",
    "get_default_columns",
    "get_default_rows",
    "prepended_sys_path",
    "resolve_builder_class",
    "resolve_builder_paths",
]



"""Tests for Colors class formatting helper methods."""

from __future__ import annotations

import pytest

from tests._internal.conftest import ANSI_BLUE, ANSI_BOLD, ANSI_CYAN, ANSI_MAGENTA
from tmuxp._internal.colors import ColorMode, Colors

# format_label tests


def test_format_label_plain_text() -> None:
    """format_label returns plain text when colors disabled."""
    colors = Colors(ColorMode.NEVER)
    assert colors.format_label("tmux path") == "tmux path"


def test_format_label_applies_highlight(monkeypatch: pytest.MonkeyPatch) -> None:
    """format_label applies highlight (bold magenta) when enabled."""
    monkeypatch.delenv("NO_COLOR", raising=False)
    colors = Colors(ColorMode.ALWAYS)

    result = colors.format_label("tmux path")
    assert ANSI_MAGENTA in result  # magenta
    assert ANSI_BOLD in result  # bold
    assert "tmux path" in result


# format_path tests


def test_format_path_plain_text() -> None:
    """format_path returns plain text when colors disabled."""
    colors = Colors(ColorMode.NEVER)
    assert colors.format_path("/usr/bin/tmux") == "/usr/bin/tmux"


def test_format_path_applies_info(monkeypatch: pytest.MonkeyPatch) -> None:
    """format_path applies info color (cyan) when enabled."""
    monkeypatch.delenv("NO_COLOR", raising=False)
    colors = Colors(ColorMode.ALWAYS)

    result = colors.format_path("/usr/bin/tmux")
    assert ANSI_CYAN in result  # cyan
    assert "/usr/bin/tmux" in result


# format_version tests


def test_format_version_plain_text() -> None:
    """format_version returns plain text when colors disabled."""
    colors = Colors(ColorMode.NEVER)
    assert colors.format_version("3.2a") == "3.2a"


def test_format_version_applies_info(monkeypatch: pytest.MonkeyPatch) -> None:
    """format_version applies info color (cyan) when enabled."""
    monkeypatch.delenv("NO_COLOR", raising=False)
    colors = Colors(ColorMode.ALWAYS)

    result = colors.format_version("3.2a")
    assert ANSI_CYAN in result  # cyan
    assert "3.2a" in result


# format_separator tests


def test_format_separator_default_length() -> None:
    """format_separator creates 25-character separator by default."""
    colors = Colors(ColorMode.NEVER)
    result = colors.format_separator()
    assert result == "-" * 25


def test_format_separator_custom_length() -> None:
    """format_separator respects custom length."""
    colors = Colors(ColorMode.NEVER)
    assert colors.format_separator(10) == "-" * 10
    assert colors.format_separator(50) == "-" * 50


def test_format_separator_applies_muted(monkeypatch: pytest.MonkeyPatch) -> None:
    """format_separator applies muted color (blue) when enabled."""
    monkeypatch.delenv("NO_COLOR", raising=False)
    colors = Colors(ColorMode.ALWAYS)

    result = colors.format_separator()
    assert ANSI_BLUE in result  # blue
    assert "-" * 25 in result


# format_kv tests


def test_format_kv_plain_text() -> None:
    """format_kv returns plain key: value when colors disabled."""
    colors = Colors(ColorMode.NEVER)
    assert colors.format_kv("tmux version", "3.2a") == "tmux version: 3.2a"


def test_format_kv_highlights_key(monkeypatch: pytest.MonkeyPatch) -> None:
    """format_kv highlights the key but not the value."""
    monkeypatch.delenv("NO_COLOR", raising=False)
    colors = Colors(ColorMode.ALWAYS)

    result = colors.format_kv("tmux version", "3.2a")
    assert ANSI_MAGENTA in result  # magenta for key
    assert ANSI_BOLD in result  # bold for key
    assert "tmux version" in result
    assert ": 3.2a" in result


def test_format_kv_empty_value() -> None:
    """format_kv handles empty value."""
    colors = Colors(ColorMode.NEVER)
    assert colors.format_kv("environment", "") == "environment: "


# format_tmux_option tests


def test_format_tmux_option_plain_text_key_value() -> None:
    """format_tmux_option returns plain text when colors disabled (key=value)."""
    colors = Colors(ColorMode.NEVER)
    assert colors.format_tmux_option("base-index=1") == "base-index=1"


def test_format_tmux_option_plain_text_space_sep() -> None:
    """format_tmux_option returns plain text when colors disabled (space-sep)."""
    colors = Colors(ColorMode.NEVER)
    assert colors.format_tmux_option("status on") == "status on"


def test_format_tmux_option_key_value_format(monkeypatch: pytest.MonkeyPatch) -> None:
    """format_tmux_option highlights key=value format."""
    monkeypatch.delenv("NO_COLOR", raising=False)
    colors = Colors(ColorMode.ALWAYS)

    result = colors.format_tmux_option("base-index=1")
    assert ANSI_MAGENTA in result  # magenta for key
    assert ANSI_CYAN in result  # cyan for value
    assert "base-index" in result
    assert "=1" in result or "1" in result


def test_format_tmux_option_space_separated(monkeypatch: pytest.MonkeyPatch) -> None:
    """format_tmux_option highlights space-separated format."""
    monkeypatch.delenv("NO_COLOR", raising=False)
    colors = Colors(ColorMode.ALWAYS)

    result = colors.format_tmux_option("status on")
    assert ANSI_MAGENTA in result  # magenta for key
    assert ANSI_CYAN in result  # cyan for value
    assert "status" in result
    assert "on" in result


def test_format_tmux_option_single_word() -> None:
    """format_tmux_option returns single words (empty array options) unchanged."""
    colors = Colors(ColorMode.NEVER)
    assert colors.format_tmux_option("pane-colours") == "pane-colours"


def test_format_tmux_option_single_word_highlighted(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """format_tmux_option highlights single words (empty array options) as keys."""
    monkeypatch.delenv("NO_COLOR", raising=False)
    colors = Colors(ColorMode.ALWAYS)

    result = colors.format_tmux_option("pane-colours")
    assert ANSI_MAGENTA in result  # magenta for key
    assert "pane-colours" in result


def test_format_tmux_option_empty() -> None:
    """format_tmux_option handles empty string."""
    colors = Colors(ColorMode.NEVER)
    assert colors.format_tmux_option("") == ""


def test_format_tmux_option_array_indexed(monkeypatch: pytest.MonkeyPatch) -> None:
    """format_tmux_option handles array-indexed keys like status-format[0]."""
    monkeypatch.delenv("NO_COLOR", raising=False)
    colors = Colors(ColorMode.ALWAYS)

    result = colors.format_tmux_option('status-format[0] "#[align=left]"')
    assert ANSI_MAGENTA in result  # magenta for key
    assert ANSI_CYAN in result  # cyan for value
    assert "status-format[0]" in result
    assert "#[align=left]" in result


def test_format_tmux_option_array_indexed_complex_value(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """format_tmux_option handles complex format strings as values."""
    monkeypatch.delenv("NO_COLOR", raising=False)
    colors = Colors(ColorMode.ALWAYS)

    # Real tmux status-format value (truncated for test)
    line = 'status-format[0] "#[align=left range=left #{E:status-left-style}]"'
    result = colors.format_tmux_option(line)
    assert "status-format[0]" in result
    assert "#[align=left" in result


def test_format_tmux_option_value_with_spaces(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """format_tmux_option handles values containing spaces."""
    monkeypatch.delenv("NO_COLOR", raising=False)
    colors = Colors(ColorMode.ALWAYS)

    # tmux options can have values with spaces like status-left "string here"
    result = colors.format_tmux_option('status-left "#S: #W"')
    assert "status-left" in result
    assert '"#S: #W"' in result


def test_format_tmux_option_value_with_equals(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """format_tmux_option splits only on first equals for key=value format."""
    monkeypatch.delenv("NO_COLOR", raising=False)
    colors = Colors(ColorMode.ALWAYS)

    # Only split on first equals (no spaces = key=value format)
    result = colors.format_tmux_option("option=a=b=c")
    assert "option" in result
    assert "a=b=c" in result
    assert ANSI_MAGENTA in result  # magenta for key
    assert ANSI_CYAN in result  # cyan for value



"""Custom help formatter for tmuxp CLI with colorized examples.

This module provides a custom argparse formatter that colorizes example
sections in help output, similar to vcspull's formatter.

Examples
--------
>>> from tmuxp.cli._formatter import TmuxpHelpFormatter
>>> TmuxpHelpFormatter  # doctest: +ELLIPSIS
<class '...TmuxpHelpFormatter'>
"""

from __future__ import annotations

import argparse
import logging
import re
import typing as t

logger = logging.getLogger(__name__)

# Options that expect a value (set externally or via --option=value)
OPTIONS_EXPECTING_VALUE = frozenset(
    {
        "-f",
        "--file",
        "-s",
        "--socket-name",
        "-S",
        "--socket-path",
        "-L",
        "--log-level",
        "-c",
        "--command",
        "-t",
        "--target",
        "-o",
        "--output",
        # Note: -d is --detached (flag-only), not a value option
        "--color",
        "-w",
        "--workspace",
    }
)

# Standalone flag options (no value)
OPTIONS_FLAG_ONLY = frozenset(
    {
        "-h",
        "--help",
        "-V",
        "--version",
        "-y",
        "--yes",
        "-n",
        "--no",
        "-d",
        "--detached",
        "-2",
        "-8",
        "-a",
        "--append",
        "--json",
        "--raw",
    }
)


class TmuxpHelpFormatter(argparse.RawDescriptionHelpFormatter):
    """Help formatter with colorized examples for tmuxp CLI.

    This formatter extends RawDescriptionHelpFormatter to preserve formatting
    of description text while adding syntax highlighting to example sections.

    The formatter uses a `_theme` attribute (set externally) to apply colors.
    If no theme is set, the formatter falls back to plain text output.

    Examples
    --------
    >>> formatter = TmuxpHelpFormatter("tmuxp")
    >>> formatter  # doctest: +ELLIPSIS
    <...TmuxpHelpFormatter object at ...>
    """

    # Theme for colorization, set by create_themed_formatter() or externally
    _theme: HelpTheme | None = None

    def _fill_text(self, text: str, width: int, indent: str) -> str:
        """Fill text, colorizing examples sections if theme is available.

        Parameters
        ----------
        text : str
            Text to format.
        width : int
            Maximum line width.
        indent : str
            Indentation prefix.

        Returns
        -------
        str
            Formatted text, with colorized examples if theme is set.

        Examples
        --------
        Without theme, returns text via parent formatter:

        >>> formatter = TmuxpHelpFormatter("test")
        >>> formatter._fill_text("hello", 80, "")
        'hello'
        """
        theme = getattr(self, "_theme", None)
        if not text or theme is None:
            return super()._fill_text(text, width, indent)

        lines = text.splitlines(keepends=True)
        formatted_lines: list[str] = []
        in_examples_block = False
        expect_value = False

        for line in lines:
            if line.strip() == "":
                in_examples_block = False
                expect_value = False
                formatted_lines.append(f"{indent}{line}")
                continue

            has_newline = line.endswith("\n")
            stripped_line = line.rstrip("\n")
            leading_length = len(stripped_line) - len(stripped_line.lstrip(" "))
            leading = stripped_line[:leading_length]
            content = stripped_line[leading_length:]
            content_lower = content.lower()
            # Recognize example section headings:
            # - "examples:" starts the examples block
            # - "X examples:" or "X:" are sub-section headings within examples
            is_examples_start = content_lower == "examples:"
            is_category_in_block = (
                in_examples_block and content.endswith(":") and not content[0].isspace()
            )
            is_section_heading = (
                content_lower.endswith("examples:") or is_category_in_block
            ) and not is_examples_start

            if is_section_heading or is_examples_start:
                formatted_content = f"{theme.heading}{content}{theme.reset}"
                in_examples_block = True
                expect_value = False
            elif in_examples_block:
                colored_content = self._colorize_example_line(
                    content,
                    theme=theme,
                    expect_value=expect_value,
                )
                expect_value = colored_content.expect_value
                formatted_content = colored_content.text
            else:
                formatted_content = stripped_line

            newline = "\n" if has_newline else ""
            formatted_lines.append(f"{indent}{leading}{formatted_content}{newline}")

        return "".join(formatted_lines)

    class _ColorizedLine(t.NamedTuple):
        """Result of colorizing an example line."""

        text: str
        expect_value: bool

    def _colorize_example_line(
        self,
        content: str,
        *,
        theme: t.Any,
        expect_value: bool,
    ) -> _ColorizedLine:
        """Colorize a single example command line.

        Parameters
        ----------
        content : str
            The line content to colorize.
        theme : Any
            Theme object with color attributes (prog, action, etc.).
        expect_value : bool
            Whether the previous token expects a value.

        Returns
        -------
        _ColorizedLine
            Named tuple with colorized text and updated expect_value state.

        Examples
        --------
        With an empty theme (no colors), returns text unchanged:

        >>> formatter = TmuxpHelpFormatter("test")
        >>> theme = HelpTheme.from_colors(None)
        >>> result = formatter._colorize_example_line(
        ...     "tmuxp load", theme=theme, expect_value=False
        ... )
        >>> result.text
        'tmuxp load'
        >>> result.expect_value
        False
        """
        parts: list[str] = []
        expecting_value = expect_value
        first_token = True
        colored_subcommand = False

        for match in re.finditer(r"\s+|\S+", content):
            token = match.group()
            if token.isspace():
                parts.append(token)
                continue

            if expecting_value:
                color = theme.label
                expecting_value = False
            elif token.startswith("--"):
                color = theme.long_option
                expecting_value = (
                    token not in OPTIONS_FLAG_ONLY and token in OPTIONS_EXPECTING_VALUE
                )
            elif token.startswith("-"):
                color = theme.short_option
                expecting_value = (
                    token not in OPTIONS_FLAG_ONLY and token in OPTIONS_EXPECTING_VALUE
                )
            elif first_token:
                color = theme.prog
            elif not colored_subcommand:
                color = theme.action
                colored_subcommand = True
            else:
                color = None

            first_token = False

            if color:
                parts.append(f"{color}{token}{theme.reset}")
            else:
                parts.append(token)

        return self._ColorizedLine(text="".join(parts), expect_value=expecting_value)


class HelpTheme(t.NamedTuple):
    """Theme colors for help output.

    Examples
    --------
    >>> from tmuxp.cli._formatter import HelpTheme
    >>> theme = HelpTheme.from_colors(None)
    >>> theme.reset
    ''
    """

    prog: str
    action: str
    long_option: str
    short_option: str
    label: str
    heading: str
    reset: str

    @classmethod
    def from_colors(cls, colors: t.Any) -> HelpTheme:
        """Create theme from a :class:`~tmuxp._internal.colors.Colors` instance.

        Parameters
        ----------
        colors : :class:`~tmuxp._internal.colors.Colors` | None
            Colors instance, or None for no colors.

        Returns
        -------
        HelpTheme
            Theme with ANSI codes if colors enabled, empty strings otherwise.

        Examples
        --------
        >>> from tmuxp.cli._colors import Colors, ColorMode
        >>> from tmuxp.cli._formatter import HelpTheme
        >>> colors = Colors(ColorMode.NEVER)
        >>> theme = HelpTheme.from_colors(colors)
        >>> theme.reset
        ''
        """
        if colors is None or not colors._enabled:
            return cls(
                prog="",
                action="",
                long_option="",
                short_option="",
                label="",
                heading="",
                reset="",
            )

        # Import style here to avoid circular import
        from tmuxp.cli._colors import style

        return cls(
            prog=style("", fg="magenta", bold=True).removesuffix("\033[0m"),
            action=style("", fg="cyan").removesuffix("\033[0m"),
            long_option=style("", fg="green").removesuffix("\033[0m"),
            short_option=style("", fg="green").removesuffix("\033[0m"),
            label=style("", fg="yellow").removesuffix("\033[0m"),
            heading=style("", fg="blue").removesuffix("\033[0m"),
            reset="\033[0m",
        )


def create_themed_formatter(
    colors: t.Any | None = None,
) -> type[TmuxpHelpFormatter]:
    """Create a help formatter class with theme bound.

    This factory creates a formatter subclass with the theme injected,
    allowing colorized help output without modifying argparse internals.

    When no colors argument is provided, uses AUTO mode which respects
    NO_COLOR, FORCE_COLOR environment variables and TTY detection.

    Parameters
    ----------
    colors : :class:`~tmuxp._internal.colors.Colors` | None
        Colors instance for styling. If None, uses
        :attr:`~tmuxp._internal.colors.ColorMode.AUTO`.

    Returns
    -------
    type[TmuxpHelpFormatter]
        Formatter class with theme bound.

    Examples
    --------
    >>> from tmuxp.cli._colors import ColorMode, Colors
    >>> from tmuxp.cli._formatter import create_themed_formatter, HelpTheme

    With explicit colors enabled:

    >>> colors = Colors(ColorMode.ALWAYS)
    >>> formatter_cls = create_themed_formatter(colors)
    >>> formatter = formatter_cls("test")
    >>> formatter._theme is not None
    True

    With colors disabled:

    >>> colors = Colors(ColorMode.NEVER)
    >>> formatter_cls = create_themed_formatter(colors)
    >>> formatter = formatter_cls("test")
    >>> formatter._theme is None
    True
    """
    # Import here to avoid circular import at module load
    from tmuxp.cli._colors import ColorMode, Colors

    if colors is None:
        colors = Colors(ColorMode.AUTO)

    # Create theme if colors are enabled, None otherwise
    theme = HelpTheme.from_colors(colors) if colors._enabled else None

    class ThemedTmuxpHelpFormatter(TmuxpHelpFormatter):
        """TmuxpHelpFormatter with theme pre-configured."""

        def __init__(self, prog: str, **kwargs: t.Any) -> None:
            super().__init__(prog, **kwargs)
            self._theme = theme

    return ThemedTmuxpHelpFormatter



"""Tests for tmuxp's utility functions."""

from __future__ import annotations

import logging
import os
import pathlib
import sys
import typing as t

import pytest

from tmuxp import exc
from tmuxp.exc import BeforeLoadScriptError, BeforeLoadScriptNotExists
from tmuxp.util import get_pane, get_session, oh_my_zsh_auto_title, run_before_script

from .constants import FIXTURE_PATH

if t.TYPE_CHECKING:
    import pathlib

    from libtmux.server import Server


def test_run_before_script_raise_BeforeLoadScriptNotExists_if_not_exists() -> None:
    """run_before_script() raises BeforeLoadScriptNotExists if script not found."""
    script_file = FIXTURE_PATH / "script_noexists.sh"

    with pytest.raises(BeforeLoadScriptNotExists):
        run_before_script(script_file)

    with pytest.raises(OSError):
        run_before_script(script_file)


def test_run_before_script_raise_BeforeLoadScriptError_if_retcode() -> None:
    """run_before_script() raises BeforeLoadScriptNotExists if script fails."""
    script_file = FIXTURE_PATH / "script_failed.sh"

    with pytest.raises(BeforeLoadScriptError):
        run_before_script(script_file)


@pytest.fixture
def temp_script(tmp_path: pathlib.Path) -> pathlib.Path:
    """Fixture of an example script that prints "Hello, world!"."""
    script = tmp_path / "test_script.sh"
    script.write_text(
        """#!/bin/sh
echo "Hello, World!"
exit 0
""",
    )
    script.chmod(0o755)
    return script


class TTYTestFixture(t.NamedTuple):
    """Test fixture for isatty behavior verification."""

    test_id: str
    isatty_value: bool
    expected_output: str


TTY_TEST_FIXTURES: list[TTYTestFixture] = [
    TTYTestFixture(
        test_id="tty_enabled_shows_output",
        isatty_value=True,
        expected_output="Hello, World!",
    ),
    TTYTestFixture(
        test_id="tty_disabled_suppresses_output",
        isatty_value=False,
        expected_output="",
    ),
]


@pytest.mark.parametrize(
    list(TTYTestFixture._fields),
    TTY_TEST_FIXTURES,
    ids=[test.test_id for test in TTY_TEST_FIXTURES],
)
def test_run_before_script_isatty(
    temp_script: pathlib.Path,
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
    test_id: str,
    isatty_value: bool,
    expected_output: str,
) -> None:
    """Verify behavior of ``isatty()``, which we mock in `run_before_script()`."""
    # Mock sys.stdout.isatty() to return the desired value.
    monkeypatch.setattr(sys.stdout, "isatty", lambda: isatty_value)

    # Run the script.
    returncode = run_before_script(temp_script)

    # Assert that the script ran successfully.
    assert returncode == 0

    out, _err = capsys.readouterr()

    # In TTY mode, we expect the output; in non-TTY mode, we expect it to be suppressed.
    assert expected_output in out


def test_return_stdout_if_ok(
    capsys: pytest.CaptureFixture[str],
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """run_before_script() returns stdout if script succeeds."""
    # Simulate sys.stdout.isatty() + sys.stderr.isatty()
    monkeypatch.setattr(sys.stdout, "isatty", lambda: True)
    monkeypatch.setattr(sys.stderr, "isatty", lambda: True)

    script_file = FIXTURE_PATH / "script_complete.sh"

    run_before_script(script_file)
    out, _err = capsys.readouterr()
    assert "hello" in out


def test_beforeload_returncode() -> None:
    """run_before_script() returns returncode if script fails."""
    script_file = FIXTURE_PATH / "script_failed.sh"

    with pytest.raises(exc.BeforeLoadScriptError) as excinfo:
        run_before_script(script_file)
        assert excinfo.match(r"113")


def test_beforeload_returns_stderr_messages() -> None:
    """run_before_script() returns stderr messages if script fails."""
    script_file = FIXTURE_PATH / "script_failed.sh"

    with pytest.raises(exc.BeforeLoadScriptError) as excinfo:
        run_before_script(script_file)
        assert excinfo.match(r"failed with returncode")


def test_get_session_should_default_to_local_attached_session(
    server: Server,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """get_session() should launch current terminal's tmux session, if inside one."""
    server.new_session(session_name="myfirstsession")
    second_session = server.new_session(session_name="mysecondsession")

    # Assign an active pane to the session
    first_pane_on_second_session_id = second_session.windows[0].panes[0].pane_id
    assert first_pane_on_second_session_id is not None
    monkeypatch.setenv("TMUX_PANE", first_pane_on_second_session_id)

    assert get_session(server) == second_session


def test_get_session_should_return_first_session_if_no_active_session(
    server: Server,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """get_session() should return first session if no active session."""
    # Clear outer tmux environment to ensure no active pane interferes
    monkeypatch.delenv("TMUX_PANE", raising=False)
    monkeypatch.delenv("TMUX", raising=False)

    first_session = server.new_session(session_name="myfirstsession")
    server.new_session(session_name="mysecondsession")

    assert get_session(server) == first_session


def test_get_pane_logs_debug_on_failure(
    server: Server,
    monkeypatch: pytest.MonkeyPatch,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """get_pane() logs DEBUG with tmux_pane extra when pane lookup fails."""
    session = server.new_session(session_name="test_pane_log")
    window = session.active_window

    # Make active_pane raise Exception to trigger the logging path
    monkeypatch.setattr(
        type(window),
        "active_pane",
        property(lambda self: (_ for _ in ()).throw(Exception("mock pane error"))),
    )

    with (
        caplog.at_level(logging.DEBUG, logger="tmuxp.util"),
        pytest.raises(exc.PaneNotFound),
    ):
        get_pane(window, current_pane=None)

    debug_records = [
        r
        for r in caplog.records
        if hasattr(r, "tmux_pane") and r.levelno == logging.DEBUG
    ]
    assert len(debug_records) >= 1
    assert debug_records[0].tmux_pane == ""


def test_oh_my_zsh_auto_title_logs_warning(
    monkeypatch: pytest.MonkeyPatch,
    caplog: pytest.LogCaptureFixture,
    tmp_path: t.Any,
) -> None:
    """oh_my_zsh_auto_title() logs WARNING when DISABLE_AUTO_TITLE not set."""
    monkeypatch.setenv("SHELL", "/bin/zsh")
    monkeypatch.delenv("DISABLE_AUTO_TITLE", raising=False)

    # Create fake ~/.oh-my-zsh directory
    fake_home = tmp_path / "home"
    fake_home.mkdir()
    oh_my_zsh_dir = fake_home / ".oh-my-zsh"
    oh_my_zsh_dir.mkdir()
    monkeypatch.setenv("HOME", str(fake_home))

    # Patch os.path.exists to return True for ~/.oh-my-zsh
    original_exists = os.path.exists

    def patched_exists(path: str) -> bool:
        if path == str(pathlib.Path("~/.oh-my-zsh").expanduser()):
            return True
        return original_exists(path)

    monkeypatch.setattr(os.path, "exists", patched_exists)

    with caplog.at_level(logging.WARNING, logger="tmuxp.util"):
        oh_my_zsh_auto_title()

    warning_records = [r for r in caplog.records if r.levelno == logging.WARNING]
    assert len(warning_records) >= 1
    assert "DISABLE_AUTO_TITLE" in warning_records[0].message



"""Public contract for tmuxp workspace builders.

A workspace builder turns an expanded workspace ``dict`` into a live tmux
session. :class:`tmuxp.workspace.builder.classic.ClassicWorkspaceBuilder` is the
default implementation; third parties may ship their own and select them with
the ``workspace_builder`` config key (see
:mod:`tmuxp.workspace.builder.registry`).

This module is intentionally dependency-light (typing only) so builder authors
can import the contract without pulling in tmuxp's resolution machinery.

The contract is synchronous today. Async builders are an *additive* future
extension: a subclass protocol can declare an ``async`` build entry point and a
capability flag, discovered through the same entry-point group, without
changing this sync surface.
"""

from __future__ import annotations

import typing as t

if t.TYPE_CHECKING:
    from libtmux.server import Server
    from libtmux.session import Session


@t.runtime_checkable
class WorkspaceBuilderProtocol(t.Protocol):
    """Contract a builder must satisfy to be driven by ``tmuxp load``.

    A conforming builder is constructed with at least ``session_config`` and
    ``server`` (plus the optional ``plugins`` and ``on_*`` callbacks the classic
    builder accepts). It exposes :meth:`build`, surfaces the populated tmux
    session via :attr:`session`, and honors the plugin/progress integration
    points the CLI drives.

    Because the protocol carries data members (``session``, ``plugins``, the
    ``on_*`` callbacks), validate *instances* with :func:`isinstance`;
    class-level ``issubclass`` checks are not supported for runtime-checkable
    protocols with non-method members.

    Examples
    --------
    A builder implementing the full contract satisfies the protocol:

    >>> from tmuxp.workspace.builder.protocol import WorkspaceBuilderProtocol
    >>> class MiniBuilder:
    ...     plugins: list = []
    ...     on_progress = None
    ...     on_before_script = None
    ...     on_script_output = None
    ...     on_build_event = None
    ...     def __init__(self, session_config, server, **kwargs):
    ...         self._session_config = session_config
    ...     def build(self, session=None, append=False):
    ...         ...
    ...     def session_exists(self, session_name):
    ...         ...
    ...     def find_current_attached_session(self):
    ...         ...
    ...     @property
    ...     def session(self):
    ...         ...
    >>> builder = MiniBuilder(session_config={}, server=None)
    >>> isinstance(builder, WorkspaceBuilderProtocol)
    True

    An object missing the contract does not:

    >>> isinstance(object(), WorkspaceBuilderProtocol)
    False
    """

    plugins: list[t.Any]
    on_progress: t.Callable[[str], None] | None
    on_before_script: t.Callable[[], None] | None
    on_script_output: t.Callable[[str], None] | None
    on_build_event: t.Callable[[dict[str, t.Any]], None] | None

    def __init__(
        self,
        session_config: dict[str, t.Any],
        server: Server,
        plugins: list[t.Any] | None = None,
        on_progress: t.Callable[[str], None] | None = None,
        on_before_script: t.Callable[[], None] | None = None,
        on_script_output: t.Callable[[str], None] | None = None,
        on_build_event: t.Callable[[dict[str, t.Any]], None] | None = None,
    ) -> None:
        """Construct a builder from an expanded workspace and a tmux server."""
        ...

    @property
    def session(self) -> Session:
        """Return the tmux session the builder created or populated."""
        ...

    def build(self, session: Session | None = None, append: bool = False) -> None:
        """Build the workspace, creating or populating a tmux session."""
        ...

    def session_exists(self, session_name: str) -> bool:
        """Return ``True`` if a session with ``session_name`` already exists."""
        ...

    def find_current_attached_session(self) -> Session:
        """Return the session currently attached within ``$TMUX``."""
        ...



"""Tests for local workspace file discovery with upward traversal."""

from __future__ import annotations

import pathlib
import typing as t

import pytest

from tmuxp.workspace.finders import LOCAL_WORKSPACE_FILES, find_local_workspace_files


class LocalWorkspaceTestFixture(t.NamedTuple):
    """Test fixture for local workspace file discovery."""

    test_id: str
    files: dict[str, str]  # {dir_relative_to_home: filename}
    start_dir: str  # relative to home
    expected_count: int
    expected_paths: list[str]  # relative to home


LOCAL_WORKSPACE_TEST_FIXTURES: list[LocalWorkspaceTestFixture] = [
    LocalWorkspaceTestFixture(
        test_id="only_in_cwd",
        files={"project": ".tmuxp.yaml"},
        start_dir="project",
        expected_count=1,
        expected_paths=["project/.tmuxp.yaml"],
    ),
    LocalWorkspaceTestFixture(
        test_id="only_in_parent",
        files={"project": ".tmuxp.yaml"},
        start_dir="project/subdir",
        expected_count=1,
        expected_paths=["project/.tmuxp.yaml"],
    ),
    LocalWorkspaceTestFixture(
        test_id="in_cwd_and_parent",
        files={
            "project": ".tmuxp.yaml",
            "project/subdir": ".tmuxp.yaml",
        },
        start_dir="project/subdir",
        expected_count=2,
        expected_paths=["project/subdir/.tmuxp.yaml", "project/.tmuxp.yaml"],
    ),
    LocalWorkspaceTestFixture(
        test_id="multiple_ancestors",
        files={
            "a": ".tmuxp.yaml",
            "a/b": ".tmuxp.yaml",
            "a/b/c": ".tmuxp.yaml",
        },
        start_dir="a/b/c/d",
        expected_count=3,
        expected_paths=[
            "a/b/c/.tmuxp.yaml",
            "a/b/.tmuxp.yaml",
            "a/.tmuxp.yaml",
        ],
    ),
    LocalWorkspaceTestFixture(
        test_id="no_local_files",
        files={},
        start_dir="project",
        expected_count=0,
        expected_paths=[],
    ),
    LocalWorkspaceTestFixture(
        test_id="json_format",
        files={"project": ".tmuxp.json"},
        start_dir="project",
        expected_count=1,
        expected_paths=["project/.tmuxp.json"],
    ),
    LocalWorkspaceTestFixture(
        test_id="yml_format",
        files={"project": ".tmuxp.yml"},
        start_dir="project",
        expected_count=1,
        expected_paths=["project/.tmuxp.yml"],
    ),
    LocalWorkspaceTestFixture(
        test_id="stops_at_home",
        files={
            "": ".tmuxp.yaml",  # In home dir itself
            "project": ".tmuxp.yaml",
        },
        start_dir="project",
        expected_count=2,  # Includes home but stops there
        expected_paths=["project/.tmuxp.yaml", ".tmuxp.yaml"],
    ),
]


@pytest.mark.parametrize(
    LocalWorkspaceTestFixture._fields,
    LOCAL_WORKSPACE_TEST_FIXTURES,
    ids=[test.test_id for test in LOCAL_WORKSPACE_TEST_FIXTURES],
)
def test_find_local_workspace_files(
    test_id: str,
    files: dict[str, str],
    start_dir: str,
    expected_count: int,
    expected_paths: list[str],
    tmp_path: pathlib.Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Test local workspace file discovery with upward traversal."""
    home = tmp_path / "home"
    home.mkdir()

    # Create directory structure and files
    for rel_dir, filename in files.items():
        dir_path = home / rel_dir if rel_dir else home
        dir_path.mkdir(parents=True, exist_ok=True)
        (dir_path / filename).write_text("session_name: test\n")

    # Ensure start directory exists
    start_path = home / start_dir
    start_path.mkdir(parents=True, exist_ok=True)

    # Mock home directory
    monkeypatch.setattr(pathlib.Path, "home", lambda: home)

    # Run the function
    result = find_local_workspace_files(start_path, stop_at_home=True)

    assert len(result) == expected_count

    # Verify paths match expected (relative to home)
    result_relative = [str(p.relative_to(home)) for p in result]
    assert result_relative == expected_paths


class TestFindLocalWorkspaceEdgeCases:
    """Edge case tests for local workspace discovery."""

    def test_at_home_directory(
        self,
        tmp_path: pathlib.Path,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        """Test behavior when starting at home directory."""
        home = tmp_path / "home"
        home.mkdir()
        monkeypatch.setattr(pathlib.Path, "home", lambda: home)

        (home / ".tmuxp.yaml").write_text("session_name: home\n")

        result = find_local_workspace_files(home, stop_at_home=True)

        assert len(result) == 1
        assert result[0] == home / ".tmuxp.yaml"

    def test_at_filesystem_root(
        self,
        tmp_path: pathlib.Path,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        """Test traversal stops at filesystem root."""
        # This test verifies no infinite loop at root
        result = find_local_workspace_files(pathlib.Path("/"), stop_at_home=False)
        # Should complete without error; result depends on system state
        assert isinstance(result, list)

    def test_yaml_precedence_over_json(
        self,
        tmp_path: pathlib.Path,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        """Test .yaml is preferred when multiple formats exist."""
        home = tmp_path / "home"
        project = home / "project"
        project.mkdir(parents=True)
        monkeypatch.setattr(pathlib.Path, "home", lambda: home)

        # Create both formats
        (project / ".tmuxp.yaml").write_text("session_name: yaml\n")
        (project / ".tmuxp.json").write_text('{"session_name": "json"}')

        result = find_local_workspace_files(project, stop_at_home=True)

        assert len(result) == 1
        assert result[0].name == ".tmuxp.yaml"

    def test_yml_precedence_over_json(
        self,
        tmp_path: pathlib.Path,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        """Test .yml is preferred when .yaml not present but .json exists."""
        home = tmp_path / "home"
        project = home / "project"
        project.mkdir(parents=True)
        monkeypatch.setattr(pathlib.Path, "home", lambda: home)

        # Create yml and json (no yaml)
        (project / ".tmuxp.yml").write_text("session_name: yml\n")
        (project / ".tmuxp.json").write_text('{"session_name": "json"}')

        result = find_local_workspace_files(project, stop_at_home=True)

        assert len(result) == 1
        assert result[0].name == ".tmuxp.yml"

    def test_stop_at_home_false_continues_past_home(
        self,
        tmp_path: pathlib.Path,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        """Test stop_at_home=False continues traversal past home."""
        # Create structure: /grandparent/home/project
        grandparent = tmp_path / "grandparent"
        home = grandparent / "home"
        project = home / "project"
        project.mkdir(parents=True)

        monkeypatch.setattr(pathlib.Path, "home", lambda: home)

        # Put config in grandparent (above home)
        (grandparent / ".tmuxp.yaml").write_text("session_name: grandparent\n")
        (project / ".tmuxp.yaml").write_text("session_name: project\n")

        # With stop_at_home=True, should only find project config
        result_stop = find_local_workspace_files(project, stop_at_home=True)
        assert len(result_stop) == 1
        assert "project" in str(result_stop[0])

        # With stop_at_home=False, should find both
        result_continue = find_local_workspace_files(project, stop_at_home=False)
        assert len(result_continue) >= 2

    def test_default_start_dir_uses_cwd(
        self,
        tmp_path: pathlib.Path,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        """Test that None start_dir uses current working directory."""
        home = tmp_path / "home"
        project = home / "project"
        project.mkdir(parents=True)
        monkeypatch.setattr(pathlib.Path, "home", lambda: home)
        monkeypatch.chdir(project)

        (project / ".tmuxp.yaml").write_text("session_name: cwd\n")

        result = find_local_workspace_files(None, stop_at_home=True)

        assert len(result) == 1
        assert result[0] == project / ".tmuxp.yaml"

    def test_symlinked_directory(
        self,
        tmp_path: pathlib.Path,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        """Test behavior with symlinked directories."""
        home = tmp_path / "home"
        real_project = home / "real_project"
        real_project.mkdir(parents=True)
        symlink_project = home / "symlink_project"
        symlink_project.symlink_to(real_project)
        monkeypatch.setattr(pathlib.Path, "home", lambda: home)

        (real_project / ".tmuxp.yaml").write_text("session_name: test\n")

        result = find_local_workspace_files(symlink_project, stop_at_home=True)

        assert len(result) == 1


class TestLocalWorkspaceFilesConstant:
    """Tests for LOCAL_WORKSPACE_FILES constant."""

    def test_constant_order(self) -> None:
        """Verify LOCAL_WORKSPACE_FILES has correct order (yaml, yml, json)."""
        assert LOCAL_WORKSPACE_FILES == [".tmuxp.yaml", ".tmuxp.yml", ".tmuxp.json"]

    def test_constant_is_list(self) -> None:
        """Verify LOCAL_WORKSPACE_FILES is a list."""
        assert isinstance(LOCAL_WORKSPACE_FILES, list)
        assert len(LOCAL_WORKSPACE_FILES) == 3



"""Tests for workspace builder options (:mod:`tmuxp.workspace.options`)."""

from __future__ import annotations

import typing as t

import pytest

from tmuxp import exc
from tmuxp.workspace.options import (
    PaneReadiness,
    WorkspaceBuilderOptions,
    resolve_session_shell,
    shell_is_zsh,
)


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (None, PaneReadiness.AUTO),
        ("auto", PaneReadiness.AUTO),
        ("AUTO", PaneReadiness.AUTO),
        (True, PaneReadiness.ALWAYS),
        ("always", PaneReadiness.ALWAYS),
        ("on", PaneReadiness.ALWAYS),
        ("yes", PaneReadiness.ALWAYS),
        ("1", PaneReadiness.ALWAYS),
        (1, PaneReadiness.ALWAYS),
        (False, PaneReadiness.NEVER),
        ("never", PaneReadiness.NEVER),
        ("off", PaneReadiness.NEVER),
        ("no", PaneReadiness.NEVER),
        ("0", PaneReadiness.NEVER),
        (0, PaneReadiness.NEVER),
        (PaneReadiness.ALWAYS, PaneReadiness.ALWAYS),
    ],
)
def test_pane_readiness_from_config(value: t.Any, expected: PaneReadiness) -> None:
    """from_config maps canonical values and truthy/falsy aliases."""
    assert PaneReadiness.from_config(value) is expected


@pytest.mark.parametrize("value", ["sometimes", "maybe", "2", ""])
def test_pane_readiness_from_config_invalid(value: str) -> None:
    """from_config rejects unknown values with an actionable error."""
    with pytest.raises(ValueError, match="pane_readiness"):
        PaneReadiness.from_config(value)


def test_workspace_builder_options_defaults() -> None:
    """An absent catalog yields AUTO readiness."""
    assert WorkspaceBuilderOptions.from_config({}).pane_readiness is PaneReadiness.AUTO
    cfg = {"session_name": "x", "windows": []}
    assert WorkspaceBuilderOptions.from_config(cfg).pane_readiness is PaneReadiness.AUTO


def test_workspace_builder_options_reads_catalog() -> None:
    """from_config reads pane_readiness from the catalog."""
    cfg = {"workspace_builder_options": {"pane_readiness": "never"}}
    options = WorkspaceBuilderOptions.from_config(cfg)
    assert options.pane_readiness is PaneReadiness.NEVER


def test_workspace_builder_options_invalid_catalog_type() -> None:
    """A non-mapping catalog is rejected as a builder-option error."""
    with pytest.raises(exc.InvalidWorkspaceBuilderOption, match="must be a mapping"):
        WorkspaceBuilderOptions.from_config({"workspace_builder_options": ["nope"]})


def test_workspace_builder_options_invalid_pane_readiness() -> None:
    """An invalid pane_readiness is wrapped as a builder-option error."""
    with pytest.raises(exc.InvalidWorkspaceBuilderOption, match="pane_readiness"):
        WorkspaceBuilderOptions.from_config(
            {"workspace_builder_options": {"pane_readiness": "sometimes"}},
        )


@pytest.mark.parametrize(
    ("shell", "expected"),
    [
        ("/usr/bin/zsh", True),
        ("/bin/zsh", True),
        ("zsh", True),
        ("/bin/bash", False),
        ("/bin/sh", False),
        ("", False),
        (None, False),
    ],
)
def test_shell_is_zsh(shell: str | None, expected: bool) -> None:
    """shell_is_zsh detects zsh by name."""
    assert shell_is_zsh(shell) is expected


class _FakeSession:
    """Minimal stand-in exposing ``show_option`` for shell resolution."""

    def __init__(self, shell: str | None) -> None:
        self._shell = shell

    def show_option(self, name: str, **kwargs: t.Any) -> str | None:
        """Return the canned default-shell value."""
        return self._shell


def test_resolve_session_shell_prefers_default_shell() -> None:
    """The tmux default-shell wins over $SHELL."""
    shell = resolve_session_shell(
        _FakeSession("/usr/bin/zsh"),
        env={"SHELL": "/bin/bash"},
    )
    assert shell == "/usr/bin/zsh"


def test_resolve_session_shell_falls_back_to_env() -> None:
    """$SHELL is the fallback when default-shell is empty."""
    shell = resolve_session_shell(_FakeSession(None), env={"SHELL": "/bin/bash"})
    assert shell == "/bin/bash"


def test_resolve_session_shell_empty_when_unknown() -> None:
    """Returns an empty string when neither source resolves."""
    assert resolve_session_shell(_FakeSession(None), env={}) == ""


class _InheritedShellSession:
    """Session whose default-shell is only visible with include_inherited."""

    def __init__(self, shell: str) -> None:
        self._shell = shell

    def show_option(
        self,
        name: str,
        *,
        include_inherited: bool = False,
        **kwargs: t.Any,
    ) -> str | None:
        """Return the shell only when inherited options are requested."""
        return self._shell if include_inherited else None


def test_resolve_session_shell_reads_inherited_default_shell() -> None:
    """A globally-inherited default-shell wins over the $SHELL fallback."""
    shell = resolve_session_shell(
        _InheritedShellSession("/usr/bin/zsh"),
        env={"SHELL": "/bin/bash"},
    )
    assert shell == "/usr/bin/zsh"



"""Tests for CLI colors in import command."""

from __future__ import annotations

import pathlib

from tests.cli.conftest import (
    ANSI_BLUE,
    ANSI_CYAN,
    ANSI_GREEN,
    ANSI_RED,
    ANSI_RESET,
)
from tmuxp._internal.private_path import PrivatePath
from tmuxp.cli._colors import Colors

# Import command color output tests


def test_import_error_unknown_format(colors_always: Colors) -> None:
    """Verify unknown format error uses error color (red)."""
    msg = "Unknown config format."
    result = colors_always.error(msg)
    assert ANSI_RED in result  # red foreground
    assert msg in result
    assert result.endswith(ANSI_RESET)  # reset at end


def test_import_success_message(colors_always: Colors) -> None:
    """Verify success messages use success color (green)."""
    result = colors_always.success("Saved to ")
    assert ANSI_GREEN in result  # green foreground
    assert "Saved to" in result


def test_import_file_path_uses_info(colors_always: Colors) -> None:
    """Verify file paths use info color (cyan)."""
    path = "/path/to/config.yaml"
    result = colors_always.info(path)
    assert ANSI_CYAN in result  # cyan foreground
    assert path in result


def test_import_muted_for_banner(colors_always: Colors) -> None:
    """Verify banner text uses muted color (blue)."""
    msg = "Configuration import does its best to convert files."
    result = colors_always.muted(msg)
    assert ANSI_BLUE in result  # blue foreground
    assert msg in result


def test_import_muted_for_separator(colors_always: Colors) -> None:
    """Verify separator uses muted color (blue)."""
    separator = "---------------------------------------------------------------"
    result = colors_always.muted(separator)
    assert ANSI_BLUE in result  # blue foreground
    assert separator in result


def test_import_colors_disabled_plain_text(colors_never: Colors) -> None:
    """Verify disabled colors return plain text."""
    assert colors_never.error("error") == "error"
    assert colors_never.success("success") == "success"
    assert colors_never.muted("muted") == "muted"
    assert colors_never.info("info") == "info"


def test_import_combined_success_format(colors_always: Colors) -> None:
    """Verify combined success + info format for 'Saved to <path>' message."""
    dest = "/home/user/.tmuxp/session.yaml"
    output = colors_always.success("Saved to ") + colors_always.info(dest) + "."
    # Should contain both green and cyan ANSI codes
    assert ANSI_GREEN in output  # green for "Saved to"
    assert ANSI_CYAN in output  # cyan for path
    assert "Saved to" in output
    assert dest in output
    assert output.endswith(".")


def test_import_help_text_with_urls(colors_always: Colors) -> None:
    """Verify help text uses muted for text and info for URLs."""
    url = "<http://tmuxp.git-pull.com/examples.html>"
    help_text = colors_always.muted(
        "tmuxp has examples in JSON and YAML format at "
    ) + colors_always.info(url)
    assert ANSI_BLUE in help_text  # blue for muted text
    assert ANSI_CYAN in help_text  # cyan for URL
    assert url in help_text


def test_import_banner_with_separator(colors_always: Colors) -> None:
    """Verify banner format with separator and instruction text."""
    config_content = "session_name: test\n"
    separator = "---------------------------------------------------------------"
    output = (
        config_content
        + colors_always.muted(separator)
        + "\n"
        + colors_always.muted("Configuration import does its best to convert files.")
        + "\n"
    )
    # Should contain blue ANSI code for muted sections
    assert ANSI_BLUE in output
    assert separator in output
    assert "Configuration import" in output
    assert config_content in output


# Privacy masking tests


def test_import_masks_home_in_save_prompt(mock_home: pathlib.Path) -> None:
    """Import should mask home directory in save prompt."""
    cwd = mock_home / "projects"
    prompt = f"Save to [{PrivatePath(cwd)}]"

    assert "[~/projects]" in prompt
    assert "/home/testuser" not in prompt


def test_import_masks_home_in_confirm_prompt(mock_home: pathlib.Path) -> None:
    """Import should mask home directory in confirmation prompt."""
    dest_path = mock_home / ".tmuxp/imported.yaml"
    prompt = f"Save to {PrivatePath(dest_path)}?"

    assert "~/.tmuxp/imported.yaml" in prompt
    assert "/home/testuser" not in prompt


def test_import_masks_home_in_saved_message(
    colors_always: Colors,
    mock_home: pathlib.Path,
) -> None:
    """Import should mask home directory in 'Saved to' message."""
    dest = mock_home / ".tmuxp/imported.yaml"
    output = (
        colors_always.success("Saved to ")
        + colors_always.info(str(PrivatePath(dest)))
        + "."
    )

    assert "~/.tmuxp/imported.yaml" in output
    assert "/home/testuser" not in output



"""CLI tests for tmuxp debug-info."""

from __future__ import annotations

import json
import typing as t

import pytest

from tmuxp import cli

if t.TYPE_CHECKING:
    import pathlib


class DebugInfoOutputFixture(t.NamedTuple):
    """Test fixture for debug-info output modes."""

    test_id: str
    args: list[str]
    expected_keys: list[str]
    is_json: bool


DEBUG_INFO_OUTPUT_FIXTURES: list[DebugInfoOutputFixture] = [
    DebugInfoOutputFixture(
        test_id="human_output_has_labels",
        args=["debug-info"],
        expected_keys=["environment", "python version", "tmux version"],
        is_json=False,
    ),
    DebugInfoOutputFixture(
        test_id="json_output_valid",
        args=["debug-info", "--json"],
        expected_keys=["environment", "python_version", "tmux_version"],
        is_json=True,
    ),
]


@pytest.mark.parametrize(
    DEBUG_INFO_OUTPUT_FIXTURES[0]._fields,
    [pytest.param(*f, id=f.test_id) for f in DEBUG_INFO_OUTPUT_FIXTURES],
)
def test_debug_info_output_modes(
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
    test_id: str,
    args: list[str],
    expected_keys: list[str],
    is_json: bool,
) -> None:
    """Test debug-info output modes (human and JSON)."""
    monkeypatch.setenv("SHELL", "/bin/bash")

    cli.cli(args)
    output = capsys.readouterr().out

    if is_json:
        data = json.loads(output)
        for key in expected_keys:
            assert key in data, f"Expected key '{key}' in JSON output"
    else:
        for key in expected_keys:
            assert key in output, f"Expected '{key}' in human output"


def test_debug_info_cli(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: pathlib.Path,
    capsys: pytest.CaptureFixture[str],
) -> None:
    """Basic CLI test for tmuxp debug-info (human output)."""
    monkeypatch.setenv("SHELL", "/bin/bash")

    cli.cli(["debug-info"])
    cli_output = capsys.readouterr().out
    assert "environment" in cli_output
    assert "python version" in cli_output
    assert "system PATH" in cli_output
    assert "tmux version" in cli_output
    assert "libtmux version" in cli_output
    assert "tmuxp version" in cli_output
    assert "tmux path" in cli_output
    assert "tmuxp path" in cli_output
    assert "shell" in cli_output
    assert "tmux session" in cli_output
    assert "tmux windows" in cli_output
    assert "tmux panes" in cli_output
    assert "tmux global options" in cli_output
    assert "tmux window options" in cli_output


def test_debug_info_json_output(
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
) -> None:
    """JSON output is valid JSON with expected structure."""
    monkeypatch.setenv("SHELL", "/bin/bash")

    cli.cli(["debug-info", "--json"])
    output = capsys.readouterr().out

    data = json.loads(output)

    # Top-level keys
    assert "environment" in data
    assert "python_version" in data
    assert "system_path" in data
    assert "tmux_version" in data
    assert "libtmux_version" in data
    assert "tmuxp_version" in data
    assert "tmux_path" in data
    assert "tmuxp_path" in data
    assert "shell" in data
    assert "tmux" in data

    # Environment structure
    env = data["environment"]
    assert "dist" in env
    assert "arch" in env
    assert "uname" in env
    assert "version" in env
    assert isinstance(env["uname"], list)

    # Tmux structure
    tmux = data["tmux"]
    assert "sessions" in tmux
    assert "windows" in tmux
    assert "panes" in tmux
    assert "global_options" in tmux
    assert "window_options" in tmux
    assert isinstance(tmux["sessions"], list)


def test_debug_info_json_no_ansi(
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
) -> None:
    """JSON output should not contain ANSI escape codes."""
    monkeypatch.setenv("SHELL", "/bin/bash")

    cli.cli(["debug-info", "--json"])
    output = capsys.readouterr().out

    # ANSI escape codes start with \x1b[ or \033[
    assert "\x1b[" not in output, "JSON output contains ANSI escape codes"
    assert "\033[" not in output, "JSON output contains ANSI escape codes"


def test_debug_info_json_paths_use_private_path(
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
) -> None:
    """JSON output should mask home directory with ~."""
    import pathlib

    # Set SHELL to a path under home directory
    shell_path = pathlib.Path.home() / ".local" / "bin" / "zsh"
    monkeypatch.setenv("SHELL", str(shell_path))

    cli.cli(["debug-info", "--json"])
    output = capsys.readouterr().out
    data = json.loads(output)

    # The shell path should be masked with ~
    assert data["shell"] == "~/.local/bin/zsh", (
        f"Expected shell path to be masked with ~, got: {data['shell']}"
    )


def test_debug_info_reports_raw_tmux_version(
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
) -> None:
    """debug-info reports tmux's raw version, keeping the point-release suffix.

    libtmux's ``get_version()`` strips the letter suffix (``"3.7a"`` -> ``"3.7"``)
    for numeric comparison; ``get_version_str()`` keeps it. Bug reports want the
    exact patch release, so debug-info reads the raw string verbatim.
    """
    monkeypatch.setenv("SHELL", "/bin/bash")
    monkeypatch.setattr(
        "tmuxp.cli.debug_info.get_version_str",
        lambda *args, **kwargs: "3.7a",
    )

    cli.cli(["debug-info", "--json"])
    output = capsys.readouterr().out
    data = json.loads(output)

    assert data["tmux_version"] == "3.7a"



"""CLI for ``tmuxp import`` subcommand."""

from __future__ import annotations

import locale
import logging
import os
import pathlib
import sys
import typing as t

from tmuxp._internal.config_reader import ConfigReader
from tmuxp._internal.private_path import PrivatePath
from tmuxp.workspace import importers
from tmuxp.workspace.finders import find_workspace_file

from ._colors import ColorMode, Colors, build_description, get_color_mode
from .utils import prompt, prompt_choices, prompt_yes_no, tmuxp_echo

logger = logging.getLogger(__name__)

IMPORT_DESCRIPTION = build_description(
    """
    Import workspaces from teamocil and tmuxinator configuration files.
    """,
    (
        (
            "teamocil",
            [
                "tmuxp import teamocil ~/.teamocil/project.yml",
            ],
        ),
        (
            "tmuxinator",
            [
                "tmuxp import tmuxinator ~/.tmuxinator/project.yml",
            ],
        ),
    ),
)

if t.TYPE_CHECKING:
    import argparse
    from typing import TypeAlias

    CLIColorModeLiteral: TypeAlias = t.Literal["auto", "always", "never"]


def get_tmuxinator_dir() -> pathlib.Path:
    """Return tmuxinator configuration directory.

    Checks for ``TMUXINATOR_CONFIG`` environmental variable.

    Returns
    -------
    pathlib.Path :
        absolute path to tmuxinator config directory

    See Also
    --------
    :func:`tmuxp.workspace.importers.import_tmuxinator`
    """
    if "TMUXINATOR_CONFIG" in os.environ:
        return pathlib.Path(os.environ["TMUXINATOR_CONFIG"]).expanduser()

    return pathlib.Path("~/.tmuxinator/").expanduser()


def get_teamocil_dir() -> pathlib.Path:
    """Return teamocil configuration directory.

    Returns
    -------
    pathlib.Path :
        absolute path to teamocil config directory

    See Also
    --------
    :func:`tmuxp.workspace.importers.import_teamocil`
    """
    return pathlib.Path("~/.teamocil/").expanduser()


def _resolve_path_no_overwrite(workspace_file: str) -> str:
    path = pathlib.Path(workspace_file).resolve()
    if path.exists():
        msg = f"{path} exists. Pick a new filename."
        raise ValueError(msg)
    return str(path)


def command_import(
    workspace_file: str,
    print_list: str,
    parser: argparse.ArgumentParser,
) -> None:
    """Import a teamocil/tmuxinator config."""


def create_import_subparser(
    parser: argparse.ArgumentParser,
) -> argparse.ArgumentParser:
    """Augment :class:`argparse.ArgumentParser` with ``import`` subparser."""
    importsubparser = parser.add_subparsers(
        title="commands",
        description="valid commands",
        help="additional help",
    )

    import_teamocil = importsubparser.add_parser(
        "teamocil",
        help="convert and import a teamocil config",
    )

    import_teamocilgroup = import_teamocil.add_mutually_exclusive_group(required=True)
    teamocil_workspace_file = import_teamocilgroup.add_argument(
        dest="workspace_file",
        type=str,
        nargs="?",
        metavar="workspace-file",
        help="checks current ~/.teamocil and current directory for yaml files",
    )
    import_teamocil.set_defaults(
        callback=command_import_teamocil,
        import_subparser_name="teamocil",
    )

    import_tmuxinator = importsubparser.add_parser(
        "tmuxinator",
        help="convert and import a tmuxinator config",
    )

    import_tmuxinatorgroup = import_tmuxinator.add_mutually_exclusive_group(
        required=True,
    )
    tmuxinator_workspace_file = import_tmuxinatorgroup.add_argument(
        dest="workspace_file",
        type=str,
        nargs="?",
        metavar="workspace-file",
        help="checks current ~/.tmuxinator and current directory for yaml files",
    )

    import_tmuxinator.set_defaults(
        callback=command_import_tmuxinator,
        import_subparser_name="tmuxinator",
    )

    try:
        import shtab

        teamocil_workspace_file.complete = shtab.FILE  # type: ignore
        tmuxinator_workspace_file.complete = shtab.FILE  # type: ignore
    except ImportError:
        pass

    return parser


class ImportConfigFn(t.Protocol):
    """Typing for import configuration callback function."""

    def __call__(self, workspace_dict: dict[str, t.Any]) -> dict[str, t.Any]:
        """Execute tmuxp import function."""
        ...


def import_config(
    workspace_file: str,
    importfunc: ImportConfigFn,
    parser: argparse.ArgumentParser | None = None,
    colors: Colors | None = None,
) -> None:
    """Import a configuration from a workspace_file."""
    if colors is None:
        colors = Colors(ColorMode.AUTO)

    existing_workspace_file = ConfigReader._from_file(pathlib.Path(workspace_file))
    cfg_reader = ConfigReader(importfunc(existing_workspace_file))

    workspace_file_format = prompt_choices(
        "Convert to",
        choices=["yaml", "json"],
        default="yaml",
        color_mode=colors.mode,
    )

    if workspace_file_format == "yaml":
        new_config = cfg_reader.dump("yaml", indent=2, default_flow_style=False)
    elif workspace_file_format == "json":
        new_config = cfg_reader.dump("json", indent=2)
    else:
        sys.exit(colors.error("Unknown config format."))

    tmuxp_echo(
        new_config
        + colors.format_separator(63)
        + "\n"
        + colors.muted("Configuration import does its best to convert files.")
        + "\n",
    )
    if prompt_yes_no(
        "The new config *WILL* require adjusting afterwards. Save config?",
        color_mode=colors.mode,
    ):
        dest = None
        while not dest:
            dest_path = prompt(
                f"Save to [{PrivatePath(os.getcwd())}]",
                value_proc=_resolve_path_no_overwrite,
                color_mode=colors.mode,
            )

            # dest = dest_prompt
            if prompt_yes_no(
                f"Save to {PrivatePath(dest_path)}?",
                color_mode=colors.mode,
            ):
                dest = dest_path

        pathlib.Path(dest).write_text(
            new_config,
            encoding=locale.getpreferredencoding(False),
        )

        logger.info(
            "workspace saved",
            extra={"tmux_config_path": str(dest)},
        )
        tmuxp_echo(
            colors.success("Saved to ") + colors.info(str(PrivatePath(dest))) + ".",
        )
    else:
        tmuxp_echo(
            colors.muted("tmuxp has examples in JSON and YAML format at ")
            + colors.info("<http://tmuxp.git-pull.com/examples.html>")
            + "\n"
            + colors.muted("View tmuxp docs at ")
            + colors.info("<http://tmuxp.git-pull.com/>"),
        )
        sys.exit()


def command_import_tmuxinator(
    workspace_file: str,
    parser: argparse.ArgumentParser | None = None,
    color: CLIColorModeLiteral | None = None,
) -> None:
    """Entrypoint for ``tmuxp import tmuxinator`` subcommand.

    Converts a tmuxinator config from workspace_file to tmuxp format and import
    it into tmuxp.
    """
    color_mode = get_color_mode(color)
    colors = Colors(color_mode)

    workspace_file = find_workspace_file(
        workspace_file,
        workspace_dir=get_tmuxinator_dir(),
    )
    import_config(workspace_file, importers.import_tmuxinator, colors=colors)


def command_import_teamocil(
    workspace_file: str,
    parser: argparse.ArgumentParser | None = None,
    color: CLIColorModeLiteral | None = None,
) -> None:
    """Entrypoint for ``tmuxp import teamocil`` subcommand.

    Convert a teamocil config from workspace_file to tmuxp format and import
    it into tmuxp.
    """
    color_mode = get_color_mode(color)
    colors = Colors(color_mode)

    workspace_file = find_workspace_file(
        workspace_file,
        workspace_dir=get_teamocil_dir(),
    )

    import_config(workspace_file, importers.import_teamocil, colors=colors)
